/*
 * shadowircd: an advanced Internet Relay Chat Daemon(ircd).
 * map.c: Sends an Undernet compatible map to a user.
 *
 * Copyright (C) 2005 William Pitcock <nenolod -at- plagiarized.info>
 *
 * ShadowIRCd is free software; for licensing conditions, see the
 * LICENSE file in the root of this source tree.
 * 
 * $Id: map.c 191 2005-04-11 19:35:42Z nenolod $
 */

/*
 * I rewrote this module to send a flattened map to those who do not
 * have the routing permission. This replaces the old behaviour of
 * displaying "you have no privs" when using /map as a non oper. It
 * also restricts the vital routing information to those on the
 * routing team.
 *
 * --nenolod 2004/11/07
 */

#include "stdinc.h"
#include "client.h"
#include "modules.h"
#include "handlers.h"
#include "numeric.h"
#include "send.h"
#include "s_conf.h"
#include "ircd.h"
#include "irc_string.h"

static void m_map(struct Client *client_p, struct Client *source_p, int parc, char *parv[]);
static void dump_map(struct Client *client_p, struct Client *root, char *pbuf);
static void flattened_map(struct Client *client_p);

struct Message map_msgtab = {
	"MAP", 0, 0, 0, 0, MFLG_SLOW, 0,
	{m_unregistered, m_map, m_ignore, m_map, m_ignore}
};

#ifndef STATIC_MODULES
void
_modinit(void)
{
	mod_add_cmd(&map_msgtab);
}

void
_moddeinit(void)
{
	mod_del_cmd(&map_msgtab);
}

const char *_version = "$Revision: 191 $";
#endif

static char buf[BUFSIZE];

/* m_map()
 *	parv[0] = sender prefix
 */
static void
m_map(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	if(IsOperRouting(source_p))
	{
		dump_map(client_p, &me, buf);
		sendto_one(client_p, form_str(RPL_MAPEND), me.name, client_p->name);
	}
	else
	{
		flattened_map(client_p);
		sendto_one(client_p, form_str(RPL_MAPEND), me.name, client_p->name);
	}
	return;
}

/* dump_map()
 *   dumps server map, called recursively.
 */
static void
dump_map(struct Client *client_p, struct Client *root_p, char *pbuf)
{
	int cnt = 0, i = 0, len;
	int users;
	dlink_node *ptr;
	struct Client *server_p;

	*pbuf = '\0';

	strncat(pbuf, root_p->name, BUFSIZE - ((size_t) pbuf - (size_t) buf));

	if(HasID(root_p) && IsOperRouting(client_p))
	{
		strncat(pbuf, "[", BUFSIZE);
		strncat(pbuf, root_p->id, BUFSIZE);
		strncat(pbuf, "]", BUFSIZE);
	}

	len = strlen(buf);
	buf[len] = ' ';

	users = dlink_list_length(&root_p->serv->users);

	if(len < 50)
	{
		for(i = len + 1; i < 50; i++)
		{
			buf[i] = '-';
		}
	}

	snprintf(buf + 50, BUFSIZE,
		 " | Users: %5lu (%4.1f%%)", dlink_list_length(&root_p->serv->users),
		 100 * (float) dlink_list_length(&root_p->serv->users) / (float) Count.total);

	sendto_one(client_p, form_str(RPL_MAP), me.name, client_p->name, buf);

	if(root_p->serv->servers.head)
	{
		cnt += dlink_list_length(&root_p->serv->servers);

		if(cnt)
		{
			if(pbuf > buf + 3)
			{
				pbuf[-2] = ' ';
				if(pbuf[-3] == '`')
					pbuf[-3] = ' ';
			}
		}
	}

	i = 1;

	DLINK_FOREACH(ptr, root_p->serv->servers.head)
	{
		server_p = ptr->data;

		*pbuf = ' ';
		if(i < cnt)
			*(pbuf + 1) = '|';
		else
			*(pbuf + 1) = '`';

		*(pbuf + 2) = '-';
		*(pbuf + 3) = ' ';
		dump_map(client_p, server_p, pbuf + 4);
		i++;
	}
}

/*
 * flattened_map - display a version of map that doesn't give away routing
 *                 information to users when flattened links is enabled.
 */
static void
flattened_map(struct Client *client_p)
{
	char buf[BUFSIZE];
	dlink_node *ptr;
	struct Client *target_p;
	int i, len;
	int cnt = 0;

	/* First display me as the root */
	strlcpy(buf, me.name, BUFSIZE);
	len = strlen(buf);
	buf[len] = ' ';

	if(len < 50)
	{
		for (i = len + 1; i < 50; i++)
		{
			buf[i] = '-';
		}
	}

	snprintf(buf + 50, BUFSIZE,
		" | Users: %5lu (%4.1f%%)", dlink_list_length(&me.serv->users),
		100 * (float) dlink_list_length(&me.serv->users) / (float) Count.total);

	sendto_one(client_p, form_str(RPL_MAP), me.name, client_p->name, buf);

	/* Next, we run through every other server and list them */
	DLINK_FOREACH(ptr, global_serv_list.head)
	{
		target_p = ptr->data;

		cnt++;

		/* Skip ourselves, it's already counted */
		if(IsMe(target_p))
			continue;

		if (cnt == dlink_list_length(&global_serv_list))
			strlcpy(buf, " `- ", BUFSIZE);
		else
			strlcpy(buf, " |- ", BUFSIZE);

		strlcat(buf, target_p->name, BUFSIZE);
		len = strlen(buf);
		buf[len] = ' ';

		if(len < 50)
		{
			for (i = len + 1; i < 50; i++)
			{
				buf[i] = '-';
			}
		}

		snprintf(buf + 50, BUFSIZE,
			" | Users: %5lu (%4.1f%%)", dlink_list_length(&target_p->serv->users),
			100 * (float) dlink_list_length(&target_p->serv->users) / (float) Count.total);

		sendto_one(client_p, form_str(RPL_MAP), me.name, client_p->name, buf);
 	}
}
