/*
 *  shadowircd: an advanced Internet Relay Chat Daemon(ircd).
 *  auth-dns.c: Functions for querying a users host.
 *
 *  Copyright (C) 2002 by the past and present ircd coders, and others.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: auth-dns.c 267 2005-04-24 04:00:58Z nenolod $
 */

#include "stdinc.h"
#include "tools.h"
#include "list.h"
#include "s_auth.h"
#include "s_conf.h"
#include "client.h"
#include "common.h"
#include "event.h"
#include "fdlist.h"		/* fdlist_add */
#include "irc_string.h"
#include "sprintf_irc.h"
#include "ircd.h"
#include "ircdauth.h"
#include "numeric.h"
#include "packet.h"
#include "irc_res.h"
#include "s_bsd.h"
#include "s_log.h"
#include "s_stats.h"
#include "send.h"
#include "memory.h"
#include "dh.h"
#include "hook.h"

extern dlink_list auth_doing_dns_list;

/*
 * auth_dns_callback - called when resolver query finishes
 * if the query resulted in a successful search, hp will contain
 * a non-null pointer, otherwise hp will be null.
 * set the client on it's way to a connection completion, regardless
 * of success of failure
 */
static void
auth_dns_callback(void *vptr, struct DNSReply *reply)
{
	struct AuthRequest *auth = (struct AuthRequest *) vptr;

	dlinkDelete(&auth->dns_node, &auth_doing_dns_list);

	if(reply != NULL)
	{
		struct sockaddr_in *v4, *v4dns;
#ifdef IPV6
		struct sockaddr_in6 *v6, *v6dns;
#endif
		int good = 1;

#ifdef IPV6
		if(auth->client->localClient->ip.ss.ss_family == AF_INET6)
		{
			v6 = (struct sockaddr_in6 *) &auth->client->localClient->ip;
			v6dns = (struct sockaddr_in6 *) &reply->addr;
			if(memcmp(&v6->sin6_addr, &v6dns->sin6_addr, sizeof(struct in6_addr)) != 0)
			{
				sendheader(auth->client, REPORT_IP_MISMATCH);
				good = 0;
			}
		}
		else
#endif
		{
			v4 = (struct sockaddr_in *) &auth->client->localClient->ip;
			v4dns = (struct sockaddr_in *) &reply->addr;
			if(v4->sin_addr.s_addr != v4dns->sin_addr.s_addr)
			{
				sendheader(auth->client, REPORT_IP_MISMATCH);
				good = 0;
			}
		}
		if(good && strlen(reply->h_name) <= HOSTLEN)
		{
			strlcpy(auth->client->host, reply->h_name, sizeof(auth->client->host));
			sendto_one(auth->client, 
				"NOTICE AUTH :*** Found your hostname: %s", 
				auth->client->host);
		}
		else if(strlen(reply->h_name) > HOSTLEN)
			sendheader(auth->client, REPORT_HOST_TOOLONG);
	}
	else
		sendheader(auth->client, REPORT_FAIL_DNS);


	MyFree(auth->client->localClient->dns_query);
	auth->client->localClient->dns_query = NULL;

	/* blow away the lock, sometimes the lock isn't killed in
	 * time.
	 */
	auth->lock = 0;

	hook_call_event("authentication_done", auth);
}

static void
start_dns_query(struct AuthRequest *auth)
{
	struct Client *client_p = auth->client;

	client_p->localClient->dns_query = MyMalloc(sizeof(struct DNSQuery));
	client_p->localClient->dns_query->ptr = auth;
	client_p->localClient->dns_query->callback = auth_dns_callback;

	sendheader(client_p, REPORT_DO_DNS);

	/* No DNS cache now, remember? -- adrian */
	gethost_byaddr(&client_p->localClient->ip, client_p->localClient->dns_query);
	SetDNSPending(auth);
	dlinkAdd(auth, &auth->dns_node, &auth_doing_dns_list);

	++auth->lock;
}

#ifndef STATIC_MODULES
void
_modinit(void)
{
	hook_add_hook("authentication_run", (hookfn *) start_dns_query);
}

void
_moddeinit(void)
{
	hook_del_hook("authentication_run", (hookfn *) start_dns_query);
}

const char *_version = "$Revision: 267 $";
#endif
