/*
 *  shadowircd: an advanced Internet Relay Chat Daemon(ircd).
 *  authentication.c: Functions for querying a users ident.
 *
 *  Copyright (C) 2002 by the past and present ircd coders, and others.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: auth-identd.c 246 2005-04-21 14:49:51Z nenolod $
 */

#include "stdinc.h"
#include "tools.h"
#include "list.h"
#include "s_auth.h"
#include "s_conf.h"
#include "client.h"
#include "common.h"
#include "event.h"
#include "fdlist.h"		/* fdlist_add */
#include "irc_string.h"
#include "sprintf_irc.h"
#include "ircd.h"
#include "ircdauth.h"
#include "numeric.h"
#include "packet.h"
#include "irc_res.h"
#include "s_bsd.h"
#include "s_log.h"
#include "s_stats.h"
#include "send.h"
#include "memory.h"
#include "dh.h"
#include "hook.h"

extern dlink_list auth_doing_ident_list;

static PF read_auth_reply;
static CNCB auth_connect_callback;

/*
 * authsenderr - handle auth send errors
 */
void
auth_error(struct AuthRequest *auth)
{
	++ServerStats->is_abad;

	fd_close(auth->fd);
	auth->fd = -1;

        if(IsAuthPending(auth))
                dlinkDelete(&auth->ident_node, &auth_doing_ident_list);
        ClearAuth(auth);

	sendheader(auth->client, REPORT_FAIL_ID);

	--auth->lock;

	hook_call_event("authentication_done", auth);
}

/*
 * start_auth_query - Flag the client to show that an attempt to 
 * contact the ident server on
 * the client's host.  The connect and subsequently the socket are all put
 * into 'non-blocking' mode.  Should the connect or any later phase of the
 * identifing process fail, it is aborted and the user is given a username
 * of "unknown".
 */
static int
start_auth_query(struct AuthRequest *auth)
{
	struct irc_ssaddr localaddr;
	socklen_t locallen = sizeof(struct irc_ssaddr);
	int fd;
#ifdef IPV6
	struct sockaddr_in6 *v6;
#else
	struct sockaddr_in *v4;
#endif

	auth->lock++;

	/* open a socket of the same type as the client socket */
	if((fd = comm_open(auth->client->localClient->ip.ss.ss_family, SOCK_STREAM,
			   0, "ident")) == -1)
	{
		report_error(L_ALL, "creating auth stream socket %s:%s",
			     get_client_name(auth->client, SHOW_IP), errno);
		ilog(L_ERROR, "Unable to create auth socket for %s",
		     get_client_name(auth->client, SHOW_IP));
		++ServerStats->is_abad;
		--auth->lock;
		return 0;
	}

	if((HARD_FDLIMIT - 10) < fd)
	{
		sendto_realops_flags(UMODE_ALL, L_ALL, "Can't allocate fd for auth on %s",
				     get_client_name(auth->client, SHOW_IP));
		fd_close(fd);
		--auth->lock;
		return 0;
	}

	sendheader(auth->client, REPORT_DO_ID);
	if(!set_non_blocking(fd))
	{
		report_error(L_ALL, NONB_ERROR_MSG, get_client_name(auth->client, SHOW_IP), errno);
		fd_close(fd);
		--auth->lock;
		return 0;
	}

	/* 
	 * get the local address of the client and bind to that to
	 * make the auth request.  This used to be done only for
	 * ifdef VIRTUAL_HOST, but needs to be done for all clients
	 * since the ident request must originate from that same address--
	 * and machines with multiple IP addresses are common now
	 */
	memset(&localaddr, 0, locallen);
	getsockname(auth->client->localClient->fd, (struct sockaddr *) &localaddr, &locallen);

#ifdef IPV6
	remove_ipv6_mapping(&localaddr);
	v6 = (struct sockaddr_in6 *) &localaddr;
	v6->sin6_port = htons(0);
#else
	localaddr.ss_len = locallen;
	v4 = (struct sockaddr_in *) &localaddr;
	v4->sin_port = htons(0);
#endif
	localaddr.ss_port = htons(0);

	auth->fd = fd;
	SetAuthConnect(auth);

	comm_connect_tcp(fd, auth->client->localClient->sockhost, 113,
			 (struct sockaddr *) &localaddr, localaddr.ss_len, auth_connect_callback,
			 auth, auth->client->localClient->ip.ss.ss_family,
			 GlobalSetOptions.ident_timeout);
	return 1;		/* We suceed here for now */
}

/*
 * GetValidIdent - parse ident query reply from identd server
 * 
 * Inputs        - pointer to ident buf
 * Output        - NULL if no valid ident found, otherwise pointer to name
 * Side effects  -
 */
/*
 * A few questions have been asked about this mess, obviously
 * it should have been commented better the first time.
 * The original idea was to remove all references to libc from ircd-hybrid.
 * Instead of having to write a replacement for sscanf(), I did a
 * rather gruseome parser here so we could remove this function call.
 * Note, that I had also removed a few floating point printfs as well (though
 * now we are still stuck with a few...)
 * Remember, we have a replacement ircd sprintf, we have bleeps fputs lib
 * it would have been nice to remove some unneeded code.
 * Oh well. If we don't remove libc stuff totally, then it would be
 * far cleaner to use sscanf()
 *
 * - Dianora
 */
static char *
GetValidIdent(char *buf)
{
	int remp = 0;
	int locp = 0;
	char *colon1Ptr;
	char *colon2Ptr;
	char *colon3Ptr;
	char *commaPtr;
	char *remotePortString;

	/* All this to get rid of a sscanf() fun. */
	remotePortString = buf;

	if((colon1Ptr = strchr(remotePortString, ':')) == NULL)
		return 0;
	*colon1Ptr = '\0';
	colon1Ptr++;

	if((colon2Ptr = strchr(colon1Ptr, ':')) == NULL)
		return 0;
	*colon2Ptr = '\0';
	colon2Ptr++;

	if((commaPtr = strchr(remotePortString, ',')) == NULL)
		return 0;
	*commaPtr = '\0';
	commaPtr++;

	if((remp = atoi(remotePortString)) == 0)
		return 0;

	if((locp = atoi(commaPtr)) == 0)
		return 0;

	/* look for USERID bordered by first pair of colons */
	if(strstr(colon1Ptr, "USERID") == NULL)
		return 0;

	if((colon3Ptr = strchr(colon2Ptr, ':')) == NULL)
		return 0;
	*colon3Ptr = '\0';
	colon3Ptr++;
	return (colon3Ptr);
}

static void
auth_connect_callback(int fd, int error, void *data)
{
	struct AuthRequest *auth = data;
	struct irc_ssaddr us;
	struct irc_ssaddr them;
	char authbuf[32];
	socklen_t ulen = sizeof(struct irc_ssaddr);
	socklen_t tlen = sizeof(struct irc_ssaddr);
	uint16_t uport, tport;
#ifdef IPV6
	struct sockaddr_in6 *v6;
#else
	struct sockaddr_in *v4;
#endif

	if(error != COMM_OK)
	{
		auth_error(auth);
		return;
	}

	if(getsockname(auth->client->localClient->fd, (struct sockaddr *) &us, (socklen_t *) & ulen)
	   || getpeername(auth->client->localClient->fd, (struct sockaddr *) &them,
			  (socklen_t *) & tlen))
	{
		ilog(L_INFO, "auth get{sock,peer}name error for %s",
		     get_client_name(auth->client, SHOW_IP));
		auth_error(auth);
		return;
	}

#ifdef IPV6
	v6 = (struct sockaddr_in6 *) &us;
	uport = ntohs(v6->sin6_port);
	v6 = (struct sockaddr_in6 *) &them;
	tport = ntohs(v6->sin6_port);
	remove_ipv6_mapping(&us);
	remove_ipv6_mapping(&them);
#else
	v4 = (struct sockaddr_in *) &us;
	uport = ntohs(v4->sin_port);
	v4 = (struct sockaddr_in *) &them;
	tport = ntohs(v4->sin_port);
	us.ss_len = ulen;
	them.ss_len = tlen;
#endif

	ircsprintf(authbuf, "%u , %u\r\n", tport, uport);

	if(send(auth->fd, authbuf, strlen(authbuf), 0) == -1)
	{
		auth_error(auth);
		return;
	}
	ClearAuthConnect(auth);
	SetAuthPending(auth);
	dlinkAddTail(auth, &auth->ident_node, &auth_doing_ident_list);
	read_auth_reply(auth->fd, auth);
}

/*
 * read_auth_reply - read the reply (if any) from the ident server 
 * we connected to.
 * We only give it one shot, if the reply isn't good the first time
 * fail the authentication entirely. --Bleep
 */
#define AUTH_BUFSIZ 128

static void
read_auth_reply(int fd, void *data)
{
	struct AuthRequest *auth = data;
	char *s = NULL;
	char *t = NULL;
	int len;
	int count;
	char buf[AUTH_BUFSIZ + 1];	/* buffer to read auth reply into */

	len = recv(auth->fd, buf, AUTH_BUFSIZ, 0);

	if(len < 0 && ignoreErrno(errno))
	{
		comm_setselect(fd, FDLIST_IDLECLIENT, COMM_SELECT_READ, read_auth_reply, auth, 0);
		return;
	}

	if(len > 0)
	{
		buf[len] = '\0';

		if((s = GetValidIdent(buf)))
		{
			t = auth->client->username;

			while(*s == '~' || *s == '^')
				s++;

			for(count = USERLEN; *s && count; s++)
			{
				if(*s == '@')
				{
					break;
				}
				if(!IsSpace(*s) && *s != ':' && *s != '[')
				{
					*t++ = *s;
					count--;
				}
			}
			*t = '\0';
		}
	}

	fd_close(auth->fd);
	auth->fd = -1;

	dlinkDelete(&auth->ident_node, &auth_doing_ident_list);

	if(s == NULL)
	{
		++ServerStats->is_abad;
		strcpy(auth->client->username, "unknown");
	}
	else
	{
		sendheader(auth->client, REPORT_FIN_ID);
		++ServerStats->is_asuc;
		SetGotId(auth->client);
	}

	hook_call_event("authentication_done", auth);
}

/*
 * delete_identd_queries()
 *
 */
void
delete_identd_queries(struct Client *target_p)
{
	dlink_node *ptr;
	dlink_node *next_ptr;
	struct AuthRequest *auth;

	DLINK_FOREACH_SAFE(ptr, next_ptr, auth_doing_ident_list.head)
	{
		auth = ptr->data;

		if(auth->client == target_p)
		{
			if(auth->fd >= 0)
			{
				fd_close(auth->fd);
				auth->fd = -1;
			}

			dlinkDelete(&auth->ident_node, &auth_doing_ident_list);
		}
	}
}

void
authentication_timeout(struct AuthRequest *auth)
{
	if(auth->fd >= 0)
		fd_close(auth->fd);

	if(IsDoingAuth(auth))
		sendheader(auth->client, REPORT_FAIL_ID);
}

#ifndef STATIC_MODULES
/* init_auth()
 *
 * Initialise the auth code
 */
void
_modinit(void)
{
	hook_add_hook("authentication_run", (hookfn *) start_auth_query);
	hook_add_hook("authentication_timeout", (hookfn *) authentication_timeout);
	hook_add_hook("delete_identd_queries", (hookfn *) delete_identd_queries);
}

void
_moddeinit(void)
{
	hook_del_hook("authentication_run", (hookfn *) start_auth_query);
	hook_del_hook("authentication_timeout", (hookfn *) authentication_timeout);
	hook_del_hook("delete_identd_queries", (hookfn *) delete_identd_queries);
}

const char *_version = "$Revision: 246 $";
#endif
