/*
 * shadowircd: an advanced internet relay chat daemon(ircd).
 * isupport.c: Routines for manipulating the 005 responses.
 *
 * Copyright (c) 2005 ShadowIRCd group
 *
 * This software is free software, see LICENSE in the root
 * of the source tree for details.
 *
 * $Id$
 */

#include "stdinc.h"
#include "defaults.h"
#include "tools.h"
#include "list.h"
#include "channel.h"
#include "channel_mode.h"
#include "client.h"
#include "modules.h"
#include "isupport.h"
#include "s_conf.h"
#include "ircd_defs.h"
#include "s_serv.h"
#include "sprintf_irc.h"
#include "numeric.h"

dlink_list ISupportEntries = { 0, 0, 0 };

/* If we have more than 10 ISupport lines, we're way too bloated. */
char *ISupportLines[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

/*
 * isupport_add: adds an isupport entity to the isupport list.
 *               if it's already in the list, bail out.
 *
 * input	- an ISupportEntry
 * output	- none
 * side effects - an ISupportEntry is added to the entries list.
 */
void
isupport_add(struct ISupportEntry *entry)
{
	dlink_node *nptr;
	struct ISupportEntry *entry_p;

	DLINK_FOREACH(nptr, ISupportEntries.head)
	{
		entry_p = nptr->data;

		if (entry_p == entry)
			return;
	}

	dlinkAddTail(entry, make_dlink_node(), &ISupportEntries);
}

/*
 * isupport_find: Finds an ISupportEntry by name.
 *
 * input	- name of the entry to locate
 * output	- the ISupportEntry being looked for or NULL if
 *		  nonexistant
 * side effects	- the ISupportEntries list is searched
 */
struct ISupportEntry *
isupport_find(char *name)
{
	dlink_node *nptr;
	struct ISupportEntry *entry;

	DLINK_FOREACH(nptr, ISupportEntries.head)
	{
		entry = nptr->data;

		if (!strcasecmp(entry->name, name))
			return entry;
	}

	return NULL;
}

/*
 * isupport_del: Deletes an ISupportEntry.
 *
 * input	- the name of the ISupportEntry to delete.
 * output	- none
 * side effects	- an ISupportEntry is deleted.
 */
void
isupport_del(char *name)
{
	dlink_node *nptr;
	struct ISupportEntry *entry;

	DLINK_FOREACH(nptr, ISupportEntries.head)
	{
		entry = nptr->data;

		if (!strcasecmp(entry->name, name))
		{
			MyFree(entry);
			dlinkDelete(nptr, &ISupportEntries);
			free_dlink_node(nptr);
		}
	}
}

/*
 * isupport_create: creates an ISupportEntry and adds it.
 *
 * input	- name, value (optional)
 * output	- nothing
 * side effects	- adds an ISupportEntry to the list.
 */
void
isupport_create(char *name, char *value)
{
	struct ISupportEntry *entry = 
		MyMalloc(sizeof(struct ISupportEntry));

	DupString(entry->name, name);
	
	if (value)
		DupString(entry->value, value);

	isupport_add(entry);
}

/*
 * generate_isupport: generates ISupportLines.
 *
 * input	- none
 * output	- none
 * side effects	- ISupportLines is flushed, then generated.
 */
void
generate_isupport()
{
	int i, tokens = 0;
	dlink_node *nptr;

	for (i = 0; ISupportLines[i]; i++)
	{
		MyFree(ISupportLines[i]);
		ISupportLines[i] = NULL;
	}

	i = 0;

	DLINK_FOREACH(nptr, ISupportEntries.head)
	{
		struct ISupportEntry *entry = nptr->data;
		char *buf = MyMalloc(TOPICLEN);

		if (!ISupportLines[i])
			ISupportLines[i] = MyMalloc(TOPICLEN);

		if (((strlen(ISupportLines[i]) + sizeof(buf)) 
			>= TOPICLEN) || tokens > 12)
		{
			tokens = 0;
			i++;
			ISupportLines[i] = MyMalloc(TOPICLEN);
		}

		if (entry->value)
			ircsprintf(buf, "%s=%s ", entry->name, entry->value);
		else
			ircsprintf(buf, "%s ", entry->name);

		strcat(ISupportLines[i], buf);

		tokens++;
	}
}

/*
 * init_isupport: Sets up the ISupport table with default values.
 *
 * input	- none
 * output	- none
 * side effects	- the ISupport table is initialized.
 */
void
init_isupport()
{
	char buf[BUFSIZE];

#ifdef HAVE_LIBCRYPTO
	isupport_create("SSL", NULL);
	isupport_create("TUNL", NULL);
#endif

	isupport_create("SMARTHELP", NULL);
	isupport_create("WALLCHOPS", NULL);
	isupport_create("ENCAP", NULL);
	isupport_create("ETRACE", NULL);
	isupport_create("TSSYNC", NULL);
	isupport_create("SVSALIAS", NULL);

	isupport_create("DCCALLOW", NULL);

	isupport_create("CALLERID", "g");
	isupport_create("USERCLOAK", "v");
	isupport_create("SAFELIST", NULL);
	isupport_create("FNC", NULL);
	isupport_create("PENALTY", NULL);
	isupport_create("KNOCK", NULL);

	isupport_create("EXCEPTS", "e");
	isupport_create("INVEX", "I");
	isupport_create("DENIES", "d");
	isupport_create("QUIETS", "q");

	isupport_create("MODES", (char *) my_itoa(MAXMODEPARAMS));

	ircsprintf(buf, "%s:%i",
		ConfigChannel.disable_local_channels ? "#" : "#&",
		ConfigChannel.max_chans_per_user);
	isupport_create("CHANLIMIT", buf);

	ircsprintf(buf, "bedqI:%i", ConfigChannel.max_bans);
	isupport_create("MAXLIST", buf);

#ifndef DISABLE_CHANOWNER
	if (ConfigFileEntry.oper_prefix)
		isupport_create("PREFIX", "(uaohv)~!@%+");
	else
		isupport_create("PREFIX", "(aohv)!@%+");
#else
	if (ConfigFileEntry.oper_prefix)
		isupport_create("PREFIX", "(uohv)~@%+");
	else
		isupport_create("PREFIX", "(ohv)@%+");
#endif

	isupport_create("CHANMODES", "bedqIa,k,l,imnpstTFczGETPScrVNOF");
	isupport_create("NETWORK", ServerInfo.network_name);
	isupport_create("CASEMAPPING", "ascii");

	isupport_create("MAXTARGETS", 
			(char *) my_itoa(ConfigFileEntry.max_targets));

	isupport_create("NICKLEN", (char *) my_itoa(NICKLEN - 1));
	isupport_create("TOPICLEN", (char *) my_itoa(TOPICLEN));
	isupport_create("KICKLEN", (char *) my_itoa(TOPICLEN));

	isupport_create("SILENCE", 
			(char *) my_itoa(ConfigFileEntry.max_silence));

	isupport_create("ELIST", "MNTCAU");
	isupport_create("LANGUAGE", (char *) get_locale());

	generate_isupport();
}

