/*
 *  shadowircd: an advanced Internet Relay Chat Daemon(ircd).
 *  modules.c: A module loader.
 *
 *  Copyright (C) 2002 by the past and present ircd coders, and others.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: modules.c 195 2005-04-12 05:21:21Z nenolod $
 */

#include "stdinc.h"
#include "tools.h"
#include "modules.h"
#include "s_log.h"
#include "ircd.h"
#include "client.h"
#include "common.h"
#include "send.h"
#include "s_conf.h"
#include "handlers.h"
#include "numeric.h"
#include "parse.h"
#include "ircd_defs.h"
#include "irc_string.h"
#include "memory.h"
#include "list.h"
#include "hook.h"

#ifdef STATIC_MODULES
extern void makevirthost(struct Client *);	/* nenolod */
#endif

#ifndef STATIC_MODULES

dlink_list modules_list = { NULL, NULL, 0 };
static char *base_autoload = NULL;

static const char *core_module_table[] = {
	"die",
	"join",
	"kick",
	"kill",
	"message",
	"mode",
	"nick",
	"part",
	"quit",
	"server",
	"sjoin",
	"squit",
	NULL
};

#define MODS_INCREMENT 10
int num_mods = 0;
int max_mods = MODS_INCREMENT;

static dlink_list mod_paths = { NULL, NULL, 0 };

static void mo_modload(struct Client *, struct Client *, int, char **);
static void mo_modlist(struct Client *, struct Client *, int, char **);
static void mo_modreload(struct Client *, struct Client *, int, char **);
static void mo_modunload(struct Client *, struct Client *, int, char **);
static void mo_modrestart(struct Client *, struct Client *, int, char **);

struct Message modload_msgtab = {
	"MODLOAD", 0, 0, 2, 0, MFLG_SLOW, 0,
	{m_unregistered, m_not_oper, m_ignore, mo_modload, m_ignore}
};

struct Message modunload_msgtab = {
	"MODUNLOAD", 0, 0, 2, 0, MFLG_SLOW, 0,
	{m_unregistered, m_not_oper, m_ignore, mo_modunload, m_ignore}
};

struct Message modreload_msgtab = {
	"MODRELOAD", 0, 0, 2, 0, MFLG_SLOW, 0,
	{m_unregistered, m_not_oper, m_ignore, mo_modreload, m_ignore}
};

struct Message modlist_msgtab = {
	"MODLIST", 0, 0, 0, 0, MFLG_SLOW, 0,
	{m_unregistered, m_not_oper, m_ignore, mo_modlist, m_ignore}
};

struct Message modrestart_msgtab = {
	"MODRESTART", 0, 0, 0, 0, MFLG_SLOW, 0,
	{m_unregistered, m_not_oper, m_ignore, mo_modrestart, m_ignore}
};


extern struct Message hash_msgtab;	/* XXX */
extern struct Message error_msgtab;

void
modules_init(void)
{
	mod_add_cmd(&modload_msgtab);
	mod_add_cmd(&modunload_msgtab);
	mod_add_cmd(&modreload_msgtab);
	mod_add_cmd(&modlist_msgtab);
	mod_add_cmd(&modrestart_msgtab);
	mod_add_cmd(&error_msgtab);
	mod_add_cmd(&hash_msgtab);
}

/* mod_find_path()
 *
 * input	- path
 * output	- none
 * side effects - returns a module path from path
 */
static struct module_path *
mod_find_path(const char *path)
{
	dlink_node *ptr;
	struct module_path *mpath;

	DLINK_FOREACH(ptr, mod_paths.head)
	{
		mpath = ptr->data;

		if(!strcmp(path, mpath->path))
			return (mpath);
	}

	return (NULL);
}

/* init_modlist()
 * Convenience function to malloc the modlist when necessary.
 *
 * input - nothing
 * output - nothing
 * side effect - modlist may have been malloc'd and modlist_valid set to 1.
 */
static void
init_modlist(void)
{
	return;
}

/* mod_set_base()
 * Defines the base path to find the official hybrid modules.
 * I.E., all core/ modules MUST be in this directory. Specified in ircd.conf.
 * 
 * input - path
 * output - none
 * side effects - sets the base path to path
 */
void
mod_set_base(void)
{
	unsigned int len;

	MyFree(base_autoload);

	len = strlen(MODPATH) + 9;
	/* whatever MODPATH + "autoload/" */

	base_autoload = MyMalloc(len + 1);
	snprintf(base_autoload, len, "%sautoload/", MODPATH);
}

/* mod_add_path()
 *
 * input	- path
 * ouput	- 
 * side effects - adds path to list
 */
void
mod_add_path(const char *path)
{
	unsigned int len = strlen(path);
	struct module_path *pathst;

	if(mod_find_path(path))
		return;

	pathst = MyMalloc(sizeof(struct module_path));
	pathst->path = MyMalloc(len + 1);

	strlcpy(pathst->path, path, len + 1);
	dlinkAdd(pathst, &pathst->node, &mod_paths);
}

/* mod_clear_paths()
 *
 * input	-
 * output	-
 * side effects - clear the lists of paths
 */
void
mod_clear_paths(void)
{
	struct module_path *pathst;
	dlink_node *ptr;
	dlink_node *next_ptr;

	DLINK_FOREACH_SAFE(ptr, next_ptr, mod_paths.head)
	{
		pathst = ptr->data;

		dlinkDelete(&pathst->node, &mod_paths);
		MyFree(pathst->path);
		MyFree(pathst);
	}
}

/* findmodule_byname
 *
 * input        - name of module to look for
 * output       - the module or NULL
 * side effects - the module list is scanned for a module.
 */
struct module *
findmodule_byname(const char *name)
{
	dlink_node *ptr;
	struct module *module_p;

	DLINK_FOREACH(ptr, modules_list.head)
	{
		module_p = ptr->data;

		if(irccmp(module_p->name, name) == 0)
			return module_p;
	}

	return NULL;
}

/* load_all_modules()
 *
 * input        - whether or not to warn on exceptions
 * output       - none
 * side effects - a directory of modules is dynamically linked into 
 *                shadow
 */
void
load_all_modules(int warn)
{
	DIR *system_module_dir = NULL;
	struct dirent *ldirent = NULL;
	char *module_fq_name;
	int len;
	unsigned int mq_len;

	modules_init();
	init_modlist();

	max_mods = MODS_INCREMENT;
	system_module_dir = opendir(base_autoload);

	if(system_module_dir == NULL)
	{
		ilog(L_WARN, "Could not load modules from %s: %s", base_autoload, strerror(errno));
		return;
	}

	while((ldirent = readdir(system_module_dir)) != NULL)
	{
		len = strlen(ldirent->d_name);

		/* On HPUX, we have *.sl as shared library extension
		 * -TimeMr14C
		 */
		if((len > 3) && (ldirent->d_name[len - 3] == '.') &&
		   (ldirent->d_name[len - 2] == 's') &&
		   ((ldirent->d_name[len - 1] == 'o') || (ldirent->d_name[len - 1] == 'l')))
		{
			if((mq_len = strlen(base_autoload) + len + 1) > PATH_MAX)
			{
				ilog(L_ERROR, "Module path for %s truncated, module not loaded!",
				     ldirent->d_name);
			}
			else	/* Guaranteed the path fits into a string of PATH_MAX now */
			{
				module_fq_name = MyMalloc(mq_len + 1);
				snprintf(module_fq_name, mq_len + 1, "%s/%s", base_autoload,
					 ldirent->d_name);
				load_a_module(module_fq_name, warn, 0);
				MyFree(module_fq_name);
			}
		}
	}

	closedir(system_module_dir);
}

/* load_core_modules()
 *
 * input        -
 * output       -
 * side effects - core modules are loaded, if any fail, kill ircd
 */
void
load_core_modules(int warn)
{
	char *module_name;
	int i;
	unsigned int m_len;

	for(i = 0; core_module_table[i]; i++)
	{
		if((m_len =
		    (strlen(MODPATH) + strlen(core_module_table[i]) + strlen(SHARED_SUFFIX) + 2)) >
		   PATH_MAX)
		{
			ilog(L_ERROR, "Path for %s%s was truncated, the module was not loaded.",
			     core_module_table[i], SHARED_SUFFIX);
		}
		else
		{
			module_name = MyMalloc(m_len + 1);
			snprintf(module_name, m_len + 1, "%s%s%s", MODPATH, core_module_table[i],
				 SHARED_SUFFIX);
			if(load_a_module(module_name, warn, 1) == -1)
			{
				ilog(L_CRIT, "Error loading core module %s%s: terminating ircd",
				     core_module_table[i], SHARED_SUFFIX);
				exit(EXIT_FAILURE);
			}
			MyFree(module_name);
		}
	}
}

/* load_one_module()
 *
 * input        -
 * output       -
 * side effects -
 */
int
load_one_module(char *path, int coremodule)
{
	char *modpath;
	dlink_node *ptr;
	struct module_path *mpath;
	struct stat statbuf;
	unsigned int m_len;

	init_modlist();

	DLINK_FOREACH(ptr, mod_paths.head)
	{
		mpath = ptr->data;

		if((m_len = strlen(mpath->path) + strlen(path) + 1) > PATH_MAX)
		{
			ilog(L_ERROR, "Path for %s/%s was truncated, not loading module from there",
			     mpath->path, path);
			continue;
		}
		else
		{
			modpath = MyMalloc(m_len + 1);
			snprintf(modpath, m_len + 1, "%s/%s", mpath->path, path);

			if((strstr(modpath, "../") == NULL) && (strstr(modpath, "/..") == NULL))
			{
				if(stat(modpath, &statbuf) == 0)
				{
					if(S_ISREG(statbuf.st_mode))
					{
						/* Regular files only please */
						return (load_a_module(modpath, 1, coremodule));
					}
				}
			}
		}
	}

	sendto_realops_flags(UMODE_ALL, L_ALL, "Cannot locate module %s", path);
	ilog(L_WARN, "Cannot locate module %s", path);
	return (-1);
}

/* load a module .. */
static void
mo_modload(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	char *m_bn;

	if(!IsAdmin(source_p))
	{
		sendto_one(source_p, form_str(ERR_NOPRIVILEGES), me.name, source_p->name);
		return;
	}

	m_bn = basename(parv[1]);

	if(findmodule_byname(m_bn) != NULL)
	{
		sendto_one(source_p, ":%s NOTICE %s :Module %s is already loaded",
			   me.name, source_p->name, m_bn);
		return;
	}

	load_one_module(parv[1], 0);
}

/* unload a module .. */
static void
mo_modunload(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	char *m_bn;
	struct module *module_p;

	if(!IsAdmin(source_p))
	{
		sendto_one(source_p, form_str(ERR_NOPRIVILEGES), me.name, source_p->name);
		return;
	}

	m_bn = basename(parv[1]);

	if((module_p = findmodule_byname(m_bn)) == NULL)
	{
		sendto_one(source_p, ":%s NOTICE %s :Module %s is not loaded",
			   me.name, source_p->name, m_bn);
		return;
	}

	if(module_p->core == 1)
	{
		sendto_one(source_p,
			   ":%s NOTICE %s :Module %s is a core module and may not be unloaded",
			   me.name, source_p->name, m_bn);
		return;
	}

	if(unload_one_module(m_bn, 1) == -1)
	{
		sendto_one(source_p, ":%s NOTICE %s :Module %s is not loaded",
			   me.name, source_p->name, m_bn);
	}
}

/* unload and load in one! */
static void
mo_modreload(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	char *m_bn;
	struct module *module_p;
	int check_core;

	if(!IsAdmin(source_p))
	{
		sendto_one(source_p, form_str(ERR_NOPRIVILEGES), me.name, source_p->name);
		return;
	}

	m_bn = basename(parv[1]);

	if((module_p = findmodule_byname(m_bn)) == NULL)
	{
		sendto_one(source_p, ":%s NOTICE %s :Module %s is not loaded",
			   me.name, source_p->name, m_bn);
		return;
	}

	check_core = module_p->core;

	if(unload_one_module(m_bn, 1) == -1)
	{
		sendto_one(source_p, ":%s NOTICE %s :Module %s is not loaded",
			   me.name, source_p->name, m_bn);
		return;
	}

	if((load_one_module(parv[1], check_core) == -1) && check_core)
	{
		sendto_realops_flags(UMODE_ALL, L_ALL, "Error reloading core "
				     "module: %s: terminating ircd", parv[1]);
		ilog(L_CRIT, "Error loading core module %s: terminating ircd", parv[1]);
		exit(0);
	}
}

/* list modules .. */
static void
mo_modlist(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	dlink_node *ptr;
	struct module *module_p;

	if(!IsAdmin(source_p))
	{
		sendto_one(source_p, form_str(ERR_NOPRIVILEGES), me.name, source_p->name);
		return;
	}

	DLINK_FOREACH(ptr, modules_list.head)
	{
		module_p = ptr->data;

		if(parc > 1)
		{
			if(match(parv[1], module_p->name))
			{
				sendto_one(source_p, form_str(RPL_MODLIST), me.name, parv[0],
					   module_p->name, module_p->address,
					   module_p->version, 
					   module_p->core ? "(core)" : "");
			}
		}
		else
		{
			sendto_one(source_p, form_str(RPL_MODLIST), me.name, parv[0],
				   module_p->name, module_p->address,
				   module_p->version, 
				   module_p->core ? "(core)" : "");
		}
	}

	sendto_one(source_p, form_str(RPL_ENDOFMODLIST), me.name, source_p->name);
}

/* unload and reload all modules */
static void
mo_modrestart(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	int modnum;
	dlink_node *ptr, *nptr;

	if(!IsAdmin(source_p))
	{
		sendto_one(source_p, form_str(ERR_NOPRIVILEGES), me.name, source_p->name);
		return;
	}

	sendto_one(source_p, ":%s NOTICE %s :Reloading all modules", me.name, source_p->name);

	modnum = modules_list.length;

	DLINK_FOREACH_SAFE(ptr, nptr, modules_list.head)
	{
		struct module *module_p = ptr->data;
		unload_one_module(module_p->name, 0);
	}

	load_all_modules(0);
	load_core_modules(0);
	rehash(0);

	sendto_realops_flags(UMODE_ALL, L_ALL,
	     "Module Restart: %d modules unloaded, %lu modules loaded",
	     modnum, modules_list.length);
	ilog(L_WARN, "Module Restart: %d modules unloaded, %lu modules loaded",
	     modnum, modules_list.length);
}

#else /* STATIC_MODULES */
#include "s_serv.h"

/* load_all_modules()
 *
 * input        -
 * output       -
 * side effects - all the msgtabs are added for static modules
 */
void
load_all_modules(int warn)
{
	mod_add_cmd(&error_msgtab);
	mod_add_cmd(&accept_msgtab);
	mod_add_cmd(&admin_msgtab);
	mod_add_cmd(&away_msgtab);
	mod_add_cmd(&tmode_msgtab);
	mod_add_cmd(&uid_msgtab);
	mod_add_cmd(&sid_msgtab);
	mod_add_cmd(&close_msgtab);
	mod_add_cmd(&connect_msgtab);
#ifdef HAVE_LIBCRYPTO
	mod_add_cmd(&challenge_msgtab);
	mod_add_cmd(&cryptlink_msgtab);
#endif
	mod_add_cmd(&die_msgtab);
	mod_add_cmd(&dmem_msgtab);
	mod_add_cmd(&eob_msgtab);
	mod_add_cmd(&info_msgtab);
	mod_add_cmd(&invite_msgtab);
	mod_add_cmd(&ison_msgtab);
	mod_add_cmd(&join_msgtab);
	mod_add_cmd(&kick_msgtab);
	mod_add_cmd(&kill_msgtab);
	mod_add_cmd(&kline_msgtab);
	mod_add_cmd(&encap_msgtab);
	mod_add_cmd(&etrace_msgtab);
	mod_add_cmd(&hash_msgtab);
	hook_add_hook("make_virthost", (hookfn *) makevirthost);
	mod_add_cmd(&ircdhelp_msgtab);
	mod_add_cmd(&ltrace_msgtab);
	mod_add_cmd(&mkpasswd_msgtab);
	mod_add_cmd(&setcloak_msgtab);
	mod_add_cmd(&setident_msgtab);
	mod_add_cmd(&setname_msgtab);
	mod_add_cmd(&shun_msgtab);
	mod_add_cmd(&silence_msgtab);
	mod_add_cmd(&svscloak_msgtab);
	mod_add_cmd(&svsident_msgtab);
	mod_add_cmd(&svsname_msgtab);
	mod_add_cmd(&svskill_msgtab);
	mod_add_cmd(&svsnick_msgtab);
#ifdef IPV6
	mod_add_cmd(&trace_msgtab4);
#endif
	mod_add_cmd(&dline_msgtab);
	mod_add_cmd(&unkline_msgtab);
	mod_add_cmd(&undline_msgtab);
	mod_add_cmd(&knock_msgtab);
	mod_add_cmd(&links_msgtab);
	mod_add_cmd(&list_msgtab);
	mod_add_cmd(&locops_msgtab);
	mod_add_cmd(&lusers_msgtab);
	mod_add_cmd(&privmsg_msgtab);
	mod_add_cmd(&notice_msgtab);
	mod_add_cmd(&mode_msgtab);
	mod_add_cmd(&motd_msgtab);
	mod_add_cmd(&names_msgtab);
	mod_add_cmd(&nick_msgtab);
	mod_add_cmd(&oper_msgtab);
	mod_add_cmd(&operwall_msgtab);
	mod_add_cmd(&part_msgtab);
	mod_add_cmd(&pass_msgtab);
	mod_add_cmd(&ping_msgtab);
	mod_add_cmd(&pong_msgtab);
	mod_add_cmd(&quit_msgtab);
	mod_add_cmd(&rehash_msgtab);
	mod_add_cmd(&restart_msgtab);
	mod_add_cmd(&resv_msgtab);
	mod_add_cmd(&server_msgtab);
	mod_add_cmd(&set_msgtab);
	mod_add_cmd(&sjoin_msgtab);
	mod_add_cmd(&squit_msgtab);
	mod_add_cmd(&stats_msgtab);
	mod_add_cmd(&svinfo_msgtab);
	mod_add_cmd(&testline_msgtab);
	mod_add_cmd(&time_msgtab);
	mod_add_cmd(&topic_msgtab);
	mod_add_cmd(&trace_msgtab);
	mod_add_cmd(&unresv_msgtab);
	mod_add_cmd(&unxline_msgtab);
	mod_add_cmd(&user_msgtab);
	mod_add_cmd(&userhost_msgtab);
	mod_add_cmd(&users_msgtab);
	mod_add_cmd(&version_msgtab);
	mod_add_cmd(&wallops_msgtab);
	mod_add_cmd(&who_msgtab);
	mod_add_cmd(&whois_msgtab);
	mod_add_cmd(&whowas_msgtab);
	mod_add_cmd(&xline_msgtab);
	mod_add_cmd(&clearchan_msgtab);
	mod_add_cmd(&help_msgtab);
	mod_add_cmd(&jupe_msgtab);
	mod_add_cmd(&killhost_msgtab);
	mod_add_cmd(&map_msgtab);
	mod_add_cmd(&ojoin_msgtab);
	mod_add_cmd(&omotd_msgtab);
	mod_add_cmd(&operspy_msgtab);
	mod_add_cmd(&tburst_msgtab);
	mod_add_cmd(&bmask_msgtab);

	mod_add_cmd(&unrestrict_msgtab);
	mod_add_cmd(&ulist_msgtab);
	mod_add_cmd(&nlist_msgtab);
	mod_add_cmd(&glist_msgtab);
	mod_add_cmd(&sendraw_msgtab);

        hook_add_hook("authentication_run", (hookfn *) start_auth_query);
        hook_add_hook("delete_identd_queries", (hookfn *) delete_identd_queries);
}
#endif /* STATIC_MODULES */
