/*
 *  shadowircd: An advanced IRC daemon (ircd).
 *  reject.c: nenolod's reject engine. (inspired from the ratbox one.)
 *
 *  $Id: reject.c 191 2005-04-11 19:35:42Z nenolod $
 */

#include "stdinc.h"
#include "client.h"
#include "s_conf.h"
#include "event.h"
#include "tools.h"
#include "reject.h"
#include "irc_string.h"
#include "common.h"
#include "memory.h"
#include "ircd.h"
#include "list.h"

dlink_list rlist;
unsigned long reject_count = 0;

struct reject_data_
{
	char banaddr[HOSTIPLEN];	/* The IP to reject. */
	time_t time;
	unsigned int count;
};

typedef struct reject_data_ reject_data;

/*
 * create_reject - adds an ip to the rejection table
 *
 * input	- the ip to reject
 * output	- the rejection object
 * side effects	- an ip is added to the rejection cache
 */
reject_data *
create_reject(char *ipaddr)
{
	reject_data *new = MyMalloc(sizeof(reject_data));
	dlink_node *entry;
	entry = make_dlink_node();

	strcpy(new->banaddr, ipaddr);

	dlinkAddTail(new, entry, &rlist);

	return new;
}

/*
 * search_reject - finds a rejection entry
 *
 * input	- address to find
 * output	- the rejection entry (if any, otherwise null)
 * side effects	- none
 */
reject_data *
search_reject(char *ipaddr)
{
	reject_data *data;
	dlink_node *entry, *next;

	DLINK_FOREACH_SAFE(entry, next, rlist.head)
	{
		data = entry->data;
		if(match(data->banaddr, ipaddr))
			return data;
	}

	return NULL;
}

/*
 * reject_expires - removes expired rejection entries from the cache
 *
 * input	- none
 * output	- none
 * side effects	- the rejection cache is scrubbed of obsolete entries
 */
void
reject_expires(void *unused)
{
	reject_data *data;
	dlink_node *entry;

	DLINK_FOREACH(entry, rlist.head)
	{
		data = (reject_data *) entry->data;
		if(data->time + 300 < CurrentTime)
		{
			dlinkDelete(entry, &rlist);
			MyFree(data);
			free_dlink_node(entry);
		}
	}
}

/*
 * init_reject - sets up the rejection system.
 *
 * input	- none
 * output	- none
 * side effects	- necessary events are added to the scheduler for rejects
 */
void
init_reject(void)
{
	eventAdd("reject_expires", reject_expires, NULL, 300);
}

/*
 * add_reject - a simple rejection handler
 *
 * input	- a Client to reject
 * output	- none
 * side effects	- a Client is added to the rejection cache
 */
void
add_reject(struct Client *client_p)
{
	reject_data *rdata;

	if((rdata = search_reject(client_p->ipaddr)) != NULL)
	{
		rdata->time = CurrentTime;
		rdata->count++;
	}
	else
	{
		rdata = create_reject(client_p->ipaddr);
		rdata->time = CurrentTime;
		rdata->count = 1;
	}
}

/*
 * check_reject - checks if a client is in the rejection cache
 *
 * input	- a client to check
 * output	- 1 if in cache, 0 if not
 * side effects - the cache is searched for a client.
 */
int
check_reject(struct Client *client_p)
{
	reject_data *rdata;

	rdata = search_reject(client_p->ipaddr);
	if(rdata != NULL)
	{
		rdata->time = CurrentTime;
		reject_count++;
		return 1;
	}

	return 0;
}

/*
 * flush_reject - flushes the reject table
 *
 * input	- none
 * output	- none
 * side effects	- the reject table is cleared out
 */
void
flush_reject(void)
{
	dlink_node *ptr, *next;
	reject_data *rdata;

	DLINK_FOREACH_SAFE(ptr, next, rlist.head)
	{
		rdata = ptr->data;
		dlinkDelete(ptr, &rlist);
		MyFree(rdata);
		free_dlink_node(ptr);
	}
}
