/*
 *  shadowircd: an advanced Internet Relay Chat Daemon(ircd).
 *  s_auth.c: Functions for querying a users ident.
 *
 *  Copyright (C) 2002 by the past and present ircd coders, and others.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: s_auth.c 253 2002-01-29 00:49:03Z nenolod $
 */

#include "stdinc.h"
#include "tools.h"
#include "list.h"
#include "s_auth.h"
#include "s_conf.h"
#include "client.h"
#include "common.h"
#include "event.h"
#include "fdlist.h"		/* fdlist_add */
#include "irc_string.h"
#include "sprintf_irc.h"
#include "ircd.h"
#include "ircdauth.h"
#include "numeric.h"
#include "packet.h"
#include "irc_res.h"
#include "s_bsd.h"
#include "s_log.h"
#include "s_stats.h"
#include "send.h"
#include "memory.h"
#include "dh.h"
#include "hook.h"

/*
 * a bit different approach
 * this replaces the original sendheader macros
 */
static const struct
{
	const char *message;
	const size_t length;
} HeaderMessages[] =
{
	/* 123456789012345678901234567890123456789012345678901234567890 */
	{"NOTICE AUTH :*** Looking up your hostname...\r\n", 46},
	{"NOTICE AUTH :*** Found your hostname\r\n", 38},
	{"NOTICE AUTH :*** Couldn't look up your hostname\r\n", 49},
	{"NOTICE AUTH :*** Checking Ident\r\n", 33},
	{"NOTICE AUTH :*** Got Ident response\r\n", 37},
	{"NOTICE AUTH :*** No Ident response\r\n", 36},
	{"NOTICE AUTH :*** Your forward and reverse DNS do not match, "
			"ignoring hostname.\r\n", 80},
	{"NOTICE AUTH :*** Your hostname is too long, ignoring hostname\r\n", 63}
};

#ifdef HAVE_LIBCRYPTO

void
sendheader(struct Client *c, int r)
{
	if(ConfigFileEntry.send_connection_headers != 0 && IsUnknown(c))
	{
		if(IsSSL(c))
		{
			fde_t *F = &fd_table[c->localClient->fd];
			if (F->flags.accept_write)
				if (SSL_accept(F->ssl) > 0)
					F->flags.accept_write = 0;
			SSL_write(F->ssl, HeaderMessages[(r)].message, HeaderMessages[(r)].length);
		}
		else
			send((c)->localClient->fd, HeaderMessages[(r)].message,
			     HeaderMessages[(r)].length, 0);
	}
}

#else
void
sendheader(struct Client *c, int r)
{
	if(ConfigFileEntry.send_connection_headers != 0 && IsUnknown(c))
		send((c)->localClient->fd, HeaderMessages[(r)].message, HeaderMessages[(r)].length,
		     0);
}
#endif

/*
 * Ok, the original was confusing.
 * Now there are two lists, an auth request can be on both at the same time
 * or only on one or the other.
 * - Dianora
 */
dlink_list auth_doing_dns_list = { NULL, NULL, 0 };
dlink_list auth_doing_ident_list = { NULL, NULL, 0 };

static EVH timeout_auth_queries_event;

/*
 * make_auth_request - allocate a new auth request
 */
static struct AuthRequest *
make_auth_request(struct Client *client)
{
	struct AuthRequest *request = (struct AuthRequest *) MyMalloc(sizeof(struct AuthRequest));
	request->fd = -1;
	request->client = client;
	request->timeout = CurrentTime + CONNECTTIMEOUT;
	request->lock = 0;
	return (request);
}

/*
 * release_auth_client - release auth client from auth system
 * this adds the client into the local client lists so it can be read by
 * the main io processing loop
 */
static void
release_auth_client(struct AuthRequest *auth)
{
	struct Client *client_p = auth->client;

	if(!client_p)
		return;

	if(auth->lock > 0)
	{
		auth->lock--;
		return;
	}

	if(!auth->lock || auth->lock < 0)
	{
		if(client_p->localClient->fd > highest_fd)
			highest_fd = client_p->localClient->fd;

		/*
		 * When a client has auth'ed, we want to start reading what 
		 * it sends us. This is what read_packet() does.
		 *     -- adrian
		 */
		if (!dlinkFind(&global_client_list, client_p))
		{
			client_p->localClient->allow_read = MAX_FLOOD;
			comm_setflush(client_p->localClient->fd, 1000,
				 flood_recalc, client_p);
			set_no_delay(client_p->localClient->fd);
			dlinkAdd(client_p, &client_p->node, &global_client_list);
			read_packet(client_p->localClient->fd, client_p);
		}
	}
}

/*
 * authsenderr - handle auth send errors
 */
void
auth_error(struct AuthRequest *auth)
{
	++ServerStats->is_abad;

	fd_close(auth->fd);
	auth->fd = -1;

        if(IsAuthPending(auth))
                dlinkDelete(&auth->ident_node, &auth_doing_ident_list);
        ClearAuth(auth);

	sendheader(auth->client, REPORT_FAIL_ID);

	if(!IsDNSPending(auth))
	{
		dlinkDelete(&auth->dns_node, &auth_doing_dns_list);
		hook_call_event("authentication_done", auth);
		MyFree(auth);
	}
}

/*
 * start_auth - starts auth (identd) and dns queries for a client
 */
void
start_auth(struct Client *client)
{
	struct AuthRequest *auth = NULL;

	if(client->flags & FLAGS_DOINGAUTH)
		return;

	assert(client != NULL);

	if(client == NULL)
		return;

	auth = make_auth_request(client);

	client->flags |= FLAGS_DOINGAUTH;

	hook_call_event("authentication_run", auth);

	if (auth->lock == 0)
		hook_call_event("authentication_done", auth);

	printf("auth->lock: %d\n", auth->lock);
}

/*
 * timeout_auth_queries - timeout resolver and identd requests
 * allow clients through if requests failed
 */
static void
timeout_auth_queries_event(void *notused)
{
	dlink_node *ptr;
	dlink_node *next_ptr;
	struct AuthRequest *auth;

	DLINK_FOREACH_SAFE(ptr, next_ptr, auth_doing_ident_list.head)
	{
		if (ptr->data)
			auth = ptr->data;
		else
		{
			dlinkDelete(ptr, &auth_doing_ident_list);
			return;
		}

		if(auth->lock == 0)
		{
			auth->client->since = CurrentTime;
			hook_call_event("authentication_done", auth);
			continue;
		}
				
		if(auth->timeout < CurrentTime)
		{
			hook_call_event("authentication_timeout", auth);

			if(IsDNSPending(auth))
			{
				struct Client *client_p = auth->client;

				ClearDNSPending(auth);
				dlinkDelete(&auth->dns_node, &auth_doing_dns_list);
				if(client_p->localClient->dns_query != NULL)
				{
					delete_resolver_queries(client_p->localClient->dns_query);
					MyFree(client_p->localClient->dns_query);
				}
				auth->client->localClient->dns_query = NULL;
				sendheader(client_p, REPORT_FAIL_DNS);
			}
			ilog(L_INFO, "DNS/AUTH timeout %s", get_client_name(auth->client, SHOW_IP));

			auth->client->since = CurrentTime;
			if(IsAuthPending(auth))
				dlinkDelete(&auth->ident_node, &auth_doing_ident_list);
			hook_call_event("authentication_done", auth);
			MyFree(auth);
		}
	}
}

/* init_auth()
 *
 * Initialise the auth code
 */
void
init_auth(void)
{
	eventAddIsh("timeout_auth_queries_event", timeout_auth_queries_event, NULL, 1);
	hook_add_hook("authentication_done", (hookfn *) release_auth_client);
}
