/*
 * shadowircd: an advanced Internet Relay Chat Daemon(ircd).
 * s_serv.c: Server related functions.
 * 
 * Copyright (C) 2002 by the past and present ircd coders, and others.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * 
 * $Id: s_serv.c 253 2002-01-29 00:49:03Z nenolod $
 */

#include "stdinc.h"
#ifdef HAVE_LIBCRYPTO
#include <openssl/rsa.h>
#include "rsa.h"
#endif
#include "tools.h"
#include "channel.h"
#include "channel_mode.h"
#include "client.h"
#include "common.h"
#include "event.h"
#include "fdlist.h"
#include "hash.h"
#include "irc_string.h"
#include "inet_misc.h"
#include "sprintf_irc.h"
#include "ircd.h"
#include "ircd_defs.h"
#include "s_bsd.h"
#include "irc_getnameinfo.h"
#include "list.h"
#include "numeric.h"
#include "packet.h"
#include "irc_res.h"
#include "s_conf.h"
#include "s_serv.h"
#include "s_log.h"
#include "s_stats.h"
#include "s_user.h"
#include "send.h"
#include "memory.h"
#include "channel.h"		/* chcap_usage_counts stuff... */
#include "hook.h"
#include "dh.h"

struct Client  *uplink = NULL;
dlink_list	cap_list = {NULL, NULL, 0};


static unsigned long freeMask;
static void	server_burst(struct Client *);
static void	burst_all(struct Client *);

static CNCB	serv_connect_callback;

/*
 * my_name_for_link() return wildcard name of my server name according to
 * given config entry --Jto
 */
const char     *
my_name_for_link(struct ConfItem *conf)
{
	struct AccessItem *aconf;

	aconf = (struct AccessItem *)map_to_conf(conf);
	if (aconf->fakename != NULL)
		return (aconf->fakename);
	else
		return (me.name);
}

/*
 * write_links_file
 */
void
write_links_file(void *notused)
{
	MessageFileLine *next_mptr = 0;
	MessageFileLine *mptr = 0;
	MessageFileLine *currentMessageLine = 0;
	MessageFileLine *newMessageLine = 0;
	MessageFile    *MessageFileptr;
	const char     *p;
	FBFILE         *file;
	char		buff      [512];
	dlink_node     *ptr;

	MessageFileptr = &ConfigFileEntry.linksfile;

	if ((file = fbopen(MessageFileptr->fileName, "w")) == NULL)
		return;

	for (mptr = MessageFileptr->contentsOfFile; mptr; mptr = next_mptr) {
		next_mptr = mptr->next;
		MyFree(mptr);
	}

	MessageFileptr->contentsOfFile = NULL;
	currentMessageLine = NULL;

	DLINK_FOREACH(ptr, global_serv_list.head) {
		size_t		nbytes = 0;
		struct Client  *target_p = ptr->data;

		/* skip ourselves, we send ourselves in /links */
		if (IsMe(target_p))
			continue;

		if (target_p->info[0])
			p = target_p->info;
		else
			p = "(Unknown Location)";

		newMessageLine = MyMalloc(sizeof(MessageFileLine));

		/*
		 * Attempt to format the file in such a way it follows the
		 * usual links output ie  "servername uplink :hops info"
		 * Mostly for aesthetic reasons - makes it look pretty in
		 * mIRC ;) - madmax
		 */

		/*
		 * For now, check this ircsprintf wont overflow - it shouldnt
		 * on a default config but it is configurable.. This should
		 * be changed to an snprintf at some point, but I'm wanting
		 * to know if this is a cause of a bug - cryogen
		 */
		assert(strlen(target_p->name) + strlen(me.name) + 6 + strlen(p) <= MESSAGELINELEN);
		ircsprintf(newMessageLine->line, "%s %s :1 %s", target_p->name, me.name, p);
		newMessageLine->next = NULL;

		if (MessageFileptr->contentsOfFile) {
			if (currentMessageLine)
				currentMessageLine->next = newMessageLine;
			currentMessageLine = newMessageLine;
		} else {
			MessageFileptr->contentsOfFile = newMessageLine;
			currentMessageLine = newMessageLine;
		}

		nbytes = ircsprintf(buff, "%s %s :1 %s\n", target_p->name, me.name, p);
		fbputs(buff, file, nbytes);
	}

	fbclose(file);
}

/*
 * hunt_server() Do the basic thing in delivering the message (command)
 * across the relays to the specific server (server) for actions.
 * 
 * Note:   The command is a format string and *MUST* be of prefixed style (e.g.
 * ":%s COMMAND %s ..."). Command can have only max 8 parameters.
 * 
 * server  parv[server] is the parameter identifying the target server.
 * 
 * *WARNING* parv[server] is replaced with the pointer to the real servername
 * from the matched client (I'm lazy now --msa).
 * 
 * returns: (see #defines)
 */
int
hunt_server(struct Client *client_p, struct Client *source_p, const char *command,
	    int server, int parc, char *parv[])
{
	struct Client  *target_p = NULL;
	struct Client  *target_tmp = NULL;
	dlink_node     *ptr;
	int		wilds;

	/*
	 * Assume it's me, if no server
	 */
	if (parc <= server || EmptyString(parv[server]) ||
	    match(me.name, parv[server]) || match(parv[server], me.name))
		return (HUNTED_ISME);

	/*
	 * These are to pickup matches that would cause the following message
	 * to go in the wrong direction while doing quick fast non-matching
	 * lookups.
	 */
	if ((target_p = find_client(parv[server])))
		if (target_p->from == source_p->from && !MyConnect(target_p))
			target_p = NULL;
	if (!target_p && (target_p = find_server(parv[server])))
		if (target_p->from == source_p->from && !MyConnect(target_p))
			target_p = NULL;

	collapse(parv[server]);
	wilds = (strchr(parv[server], '?') || strchr(parv[server], '*'));

	/*
	 * Again, if there are no wild cards involved in the server name, use
	 * the hash lookup
	 */
	if (target_p == NULL) {
		if (!wilds) {
			if (!(target_p = find_server(parv[server]))) {
				sendto_one(source_p, form_str(ERR_NOSUCHSERVER),
					   me.name, parv[0], parv[server]);
				return (HUNTED_NOSUCH);
			}
		} else {
			DLINK_FOREACH_PREV(ptr, global_client_list.tail) {
				target_tmp = ptr->data;

				if (match(parv[server], target_tmp->name)) {
					if (target_tmp->from == source_p->from
					    && !MyConnect(target_tmp))
						continue;
					target_p = ptr->data;

					if (IsRegistered(target_p) && (target_p != client_p))
						break;
				}
			}
		}
	}
	if (target_p != NULL) {
		if (!IsRegistered(target_p)) {
			sendto_one(source_p, form_str(ERR_NOSUCHSERVER),
				   me.name, parv[0], parv[server]);
			return HUNTED_NOSUCH;
		}
		if (IsMe(target_p) || MyClient(target_p))
			return HUNTED_ISME;

		if (!match(target_p->name, parv[server]))
			parv[server] = target_p->name;

		/* This is a little kludgy but should work... */
		if (IsClient(source_p) &&
		    ((MyConnect(target_p) && ((target_p->protocol > PROTO_MIN))) ||
		     (!MyConnect(target_p) && (target_p->from->protocol > PROTO_MIN))))
			parv[0] = ID(source_p);

		sendto_one(target_p, command, parv[0],
			   parv[1], parv[2], parv[3], parv[4], parv[5], parv[6], parv[7], parv[8]);
		return (HUNTED_PASS);
	}
	sendto_one(source_p, form_str(ERR_NOSUCHSERVER), me.name, parv[0], parv[server]);
	return (HUNTED_NOSUCH);
}

/*
 * try_connections() scan through configuration and try new connections.
 * Returns the calendar time when the next call to this function should be
 * made latest. (No harm done if this is called earlier or later...)
 */
void
try_connections(void *unused)
{
	struct Client  *pending_connection;
	dlink_node     *ptr;
	struct ConfItem *conf;
	struct AccessItem *aconf;
	int		confrq;

	/*
	 * TODO: change this to set active flag to 0 when added to event!
	 * --Habeeb
	 */
	if (GlobalSetOptions.autoconn == 0)
		return;

	DLINK_FOREACH(ptr, server_items.head) {
		conf = ptr->data;
		aconf = (struct AccessItem *)map_to_conf(conf);

		/*
		 * Also when already connecting! (update holdtimes) --SRB
		 */
		if (!(aconf->status & CONF_SERVER) || aconf->port <= 0 ||
		    !(IsConfAllowAutoConn(aconf)))
			continue;

		if (aconf->hold > RealTime)
			continue;

		confrq = DEFAULT_CONNECTFREQUENCY;
		aconf->hold = RealTime + confrq;

		/*
		 * Found a CONNECT config with port specified, scan clients
		 * and see if this server is already connected?
		 */
		if (find_server(conf->name) != NULL)
			continue;

		/* Go to the end of the list, if not already last */
		if (ptr->next != NULL) {
			dlinkDelete(ptr, &server_items);
			dlinkAddTail(conf, &conf->node, &server_items);
		}

		DLINK_FOREACH(ptr, unknown_list.head) {
			pending_connection = ptr->data;

			if (pending_connection->name[0] && conf->name != NULL) {
				if (0 == irccmp(conf->name, pending_connection->name))
					return;
			}
		}

		/*
		 * We used to only print this if serv_connect()
		 * actually succeeded, but since comm_tcp_connect()
		 * can call the callback immediately if there is an
		 * error, we were getting error messages in the wrong
		 * order. SO, we just print out the activated line,
		 * and let serv_connect() / serv_connect_callback()
		 * print an error afterwards if it fails. -- adrian
		 */
		sendto_realops_flags(UMODE_ALL, L_ALL, "Connection to %s[%s] activated.",
				     conf->name, aconf->host);

		serv_connect(aconf, NULL);
		/* We connect only one at time... */
		return;
	}
}

int
check_server(const char *name, struct Client *client_p, int cryptlink)
{
	dlink_node     *ptr;
	struct ConfItem *conf = NULL;
	struct ConfItem *server_conf = NULL;
	struct AccessItem *server_aconf = NULL;
	struct AccessItem *aconf = NULL;
	int		error = -1;

	assert(client_p != NULL);

	if (client_p == NULL)
		return (error);

	if (strlen(name) > HOSTLEN)
		return (-4);

	/*
	 * loop through looking for all possible connect items that might
	 * work
	 */
	DLINK_FOREACH(ptr, server_items.head) {
		conf = ptr->data;
		aconf = (struct AccessItem *)map_to_conf(conf);

		if (!match(name, conf->name))
			continue;

		error = -3;

		/* XXX: Fix me for IPv6                    */
		/* XXX sockhost is the IPv4 ip as a string */
		if (match(aconf->host, client_p->host) ||
		    match(aconf->host, client_p->localClient->sockhost)) {
			error = -2;
			if (EmptyString(client_p->localClient->passwd))
				return (-2);

			if (IsConfEncrypted(aconf)) {
				if (strcmp(aconf->passwd,
					   (const char *)crypt(client_p->localClient->
				       passwd, aconf->passwd)) == 0)
					server_conf = conf;
			} else {
				if (strcmp(aconf->passwd, client_p->localClient->passwd) ==
				    0)
					server_conf = conf;
			}
		}
	}

	if (server_conf == NULL)
		return (error);

	attach_conf(client_p, server_conf);

	/* Now find all leaf or hub config items for this server */
	DLINK_FOREACH(ptr, hub_items.head) {
		conf = ptr->data;

		if (!match(name, conf->name))
			continue;
		attach_conf(client_p, conf);
	}

	DLINK_FOREACH(ptr, leaf_items.head) {
		conf = ptr->data;

		if (!match(name, conf->name))
			continue;
		attach_conf(client_p, conf);
	}

	server_aconf = (struct AccessItem *)map_to_conf(server_conf);

#ifdef HAVE_LIBZ		/* otherwise, clear it unconditionally */
	if (!(server_aconf->flags & CONF_FLAGS_COMPRESSED))
#endif
		client_p->ziplink = 0;

	if (aconf != NULL) {
		struct sockaddr_in *v4;
#ifdef IPV6
		struct sockaddr_in6 *v6;
#endif
		switch (aconf->aftype) {
#ifdef IPV6
		case AF_INET6:
			v6 = (struct sockaddr_in6 *)&aconf->ipnum;

			if (IN6_IS_ADDR_UNSPECIFIED(&v6->sin6_addr))
				memcpy(&aconf->ipnum, &client_p->localClient->ip,
				       sizeof(struct irc_ssaddr));
			break;
#endif
		case AF_INET:
			v4 = (struct sockaddr_in *)&aconf->ipnum;

			if (v4->sin_addr.s_addr == INADDR_NONE)
				memcpy(&aconf->ipnum, &client_p->localClient->ip,
				       sizeof(struct irc_ssaddr));
			break;
		}
	}
	return (0);
}

/*
 * add_capability()
 * 
 * inputs	- string name of CAPAB - int flag of capability output	- NONE side
 * effects	-
 * 
 */
void
add_capability(const char *capab_name, int cap_flag, int add_to_default)
{
	struct Capability *cap;

	cap = (struct Capability *)MyMalloc(sizeof(*cap));
	DupString(cap->name, capab_name);
	cap->cap = cap_flag;
	dlinkAdd(cap, &cap->node, &cap_list);
	if (add_to_default)
		default_server_capabs |= cap_flag;
}

/*
 * delete_capability()
 * 
 * inputs	- string name of CAPAB output	- NONE side effects	-
 * 
 */
int
delete_capability(const char *capab_name)
{
	dlink_node     *ptr;
	dlink_node     *next_ptr;
	struct Capability *cap;

	DLINK_FOREACH_SAFE(ptr, next_ptr, cap_list.head) {
		cap = ptr->data;

		if (cap->cap != 0) {
			if (irccmp(cap->name, capab_name) == 0) {
				default_server_capabs &= ~(cap->cap);
				dlinkDelete(ptr, &cap_list);
				MyFree(cap->name);
				cap->name = NULL;
				MyFree(cap);
			}
		}
	}

	return (0);
}

/*
 * find_capability()
 * 
 * inputs	- string name of capab to find output	- 0 if not found CAPAB
 * otherwise side effects	- none
 */
int
find_capability(const char *capab)
{
	dlink_node     *ptr;
	struct Capability *cap;

	DLINK_FOREACH(ptr, cap_list.head) {
		cap = ptr->data;

		if (cap->cap != 0) {
			if (irccmp(cap->name, capab) == 0)
				return (cap->cap);
		}
	}
	return (0);
}

/*
 * send_capabilities()
 * 
 * inputs	- Client pointer to send to - int flag of capabilities that this
 * server has output	- NONE side effects	- send the CAPAB line to a
 * server  -orabidoo
 * 
 */
void
send_capabilities(struct Client *client_p, struct AccessItem *aconf,
		  int cap_can_send, int enc_can_send)
{
	struct Capability *cap = NULL;
	char		msgbuf    [BUFSIZE];
	char           *t;
	int		tl;
	dlink_node     *ptr;
#if 0
	struct EncCapability *epref;
	char           *capend;
	int		sent_cipher = 0;
#endif

	t = msgbuf;

	DLINK_FOREACH(ptr, cap_list.head) {
		cap = ptr->data;

		if (cap->cap & (cap_can_send | default_server_capabs)) {
			tl = ircsprintf(t, "%s ", cap->name);
			t += tl;
		}
	}
#if 0
	if (enc_can_send) {
		capend = t;
		strcpy(t, "ENC:");
		t += 4;

		/* use connect{} specific info if available */
		if (aconf->cipher_preference)
			epref = aconf->cipher_preference;
		else
			epref = ConfigFileEntry.default_cipher_preference;

		if ((epref->cap & enc_can_send)) {
			/* Leave the space -- it is removed later. */
			tl = ircsprintf(t, "%s ", epref->name);
			t += tl;
			sent_cipher = 1;
		}
		if (!sent_cipher)
			t = capend;	/* truncate string before ENC:, below */
	}
#endif
	t--;
	*t = '\0';
	sendto_one(client_p, "CAPAB :%s", msgbuf);
}

/*
 * sendnick_TS()
 * 
 * inputs	- client (server) to send nick towards - client to send nick for
 * output	- NONE side effects	- NICK message is sent towards given
 * client_p
 */
void
sendnick_TS(struct Client *client_p, struct Client *target_p)
{
	if (!IsPerson(target_p))
		return;

	/* XXX Both of these need to have a :me.name or :mySID!?!?! */
	if (HasID(target_p) && client_p->protocol > PROTO_MIN)
		sendto_one(client_p, ":%s UID %s %d %lu +%s %s %s %s %s %s :%s",
			   target_p->user->server->id,
			   target_p->name, target_p->hopcount + 1,
			   (unsigned long)target_p->tsinfo,
			   umodes_as_string(&target_p->umodes), target_p->username, target_p->host,
			   target_p->ipaddr, target_p->id, target_p->virthost, target_p->info);
	else
		sendto_one(client_p, "NICK %s %d %lu +%s %s %s %s :%s",
			   target_p->name, target_p->hopcount + 1,
			   (unsigned long)target_p->tsinfo,
			   umodes_as_string(&target_p->umodes), target_p->username, target_p->host,
			   target_p->user->server->name, target_p->info);

	if (client_p->protocol == PROTO_MIN)
		sendto_one(client_p, ":%s SVSCLOAK %s :%s",
			   me.name, target_p->name, target_p->virthost);
}

/*
 * show_capabilities - show current server capabilities
 * 
 * inputs       - pointer to an struct Client output       - pointer to static
 * string side effects - build up string representing capabilities of server
 * listed
 */
const char     *
show_capabilities(struct Client *target_p)
{
	static char	msgbuf[BUFSIZE];
	struct Capability *cap;
	char           *t;
	int		tl;
	dlink_node     *ptr;

	t = msgbuf;
	tl = ircsprintf(msgbuf, "TS ");
	t += tl;

	if (!target_p->localClient->caps) {	/* short circuit if no caps */
		msgbuf[2] = '\0';
		return (msgbuf);
	}
	DLINK_FOREACH(ptr, cap_list.head) {
		cap = ptr->data;

		if (cap->cap & target_p->localClient->caps) {
			tl = ircsprintf(t, "%s ", cap->name);
			t += tl;
		}
	}
#if 0
	if (IsCapable(target_p, CAP_ENC) &&
	    target_p->localClient->in_cipher && target_p->localClient->out_cipher) {
		tl = ircsprintf(t, "ENC:%s ", target_p->localClient->in_cipher->name);
		t += tl;
	}
#endif
	t--;
	*t = '\0';

	return (msgbuf);
}

/*
 * server_estab()
 * 
 * inputs       - pointer to a struct Client output       - side effects -
 */
int
server_estab(struct Client *client_p)
{
	struct Client  *target_p;
	struct ConfItem *conf;
	struct AccessItem *aconf = NULL;
	char           *host;
	const char     *inpath;
	static char	inpath_ip[HOSTLEN * 2 + USERLEN + 6];
	dlink_node     *m;
	dlink_node     *ptr;

	assert(client_p != NULL);

	if (client_p == NULL)
		return (-1);

	strlcpy(inpath_ip, get_client_name(client_p, SHOW_IP), sizeof(inpath_ip));

	inpath = get_client_name(client_p, MASK_IP);	/* "refresh" inpath with
							 * host */
	host = client_p->name;

	if ((conf = find_conf_name(&client_p->localClient->confs, host, SERVER_TYPE)) == NULL) {
		/* This shouldn't happen, better tell the ops... -A1kmm */
		sendto_realops_flags(UMODE_ALL, L_ALL, "Warning: Lost connect{} block "
			     "for server %s(this shouldn't happen)!", host);
		return (exit_client(client_p, client_p, client_p, "Lost connect{} block!"));
	}
	MyFree(client_p->localClient->passwd);
	client_p->localClient->passwd = NULL;

	/* Its got identd , since its a server */
	SetGotId(client_p);

	/*
	 * If there is something in the serv_list, it might be this
	 * connecting server..
	 */
	if (!ServerInfo.hub && serv_list.head) {
		if (client_p != serv_list.head->data || serv_list.head->next) {
			ServerStats->is_ref++;
			sendto_one(client_p, "ERROR :I'm a leaf not a hub");
			return (exit_client(client_p, client_p, client_p, "I'm a leaf"));
		}
	}
	aconf = (struct AccessItem *)map_to_conf(conf);

	if (IsUnknown(client_p)) {
		/*
		 * jdc -- 1.  Use EmptyString(), not [0] index reference. 2.
		 * Check aconf->spasswd, not aconf->passwd.
		 */
		if (!EmptyString(aconf->spasswd)) {
			/* only send ts6 format PASS if we have ts6 enabled */
			if (me.id[0])
				sendto_one(client_p, "PASS %s TS %d %s",
					 aconf->spasswd, TS_CURRENT, me.id);
			else
				sendto_one(client_p, "PASS %s TS 5", aconf->spasswd);
		}
		sendto_one(client_p, "SERVER %s 1 %d %s :%s",
			   my_name_for_link(conf), PROTO_CUR,
			   aconf->flags & CONF_FLAGS_COMPRESSED ? "ZIP" : "",
			   (me.info[0]) ? (me.info) : "IRCers United");
	}
	/* XXX - this should be in s_bsd */
	if (!set_sock_buffers(client_p->localClient->fd, READBUF_SIZE))
		report_error(L_ALL, SETBUF_ERROR_MSG, get_client_name(client_p, SHOW_IP), errno);

	sendto_one(client_p, "SVINFO %d 3 0 :%lu",
		   TS_CURRENT, RealTime);

	/*
	 * assumption here is if they passed the correct TS version, they
	 * also passed an SID
	 */
	if (client_p->protocol > PROTO_MIN)
		hash_add_id(client_p);

	/* XXX Does this ever happen? I don't think so -db */
	detach_conf(client_p, OPER_TYPE);

	/*
	 * *WARNING* *    In the following code in place of plain server's *
	 * name we send what is returned by get_client_name *    which may
	 * add the "sockhost" after the name. It's *    *very* *important*
	 * that there is a SPACE between *    the name and sockhost (if
	 * present). The receiving *    server will start the information
	 * field from this *    first blank and thus puts the sockhost into
	 * info. *    ...a bit tricky, but you have been warned, besides *
	 * code is more neat this way...  --msa
	 */
	client_p->servptr = &me;

	if (IsClosing(client_p))
		return (CLIENT_EXITED);

	SetServer(client_p);

	/* Some day, all these lists will be consolidated *sigh* */
	dlinkAdd(client_p, &client_p->lnode, &me.serv->servers);

	m = dlinkFind(&unknown_list, client_p);
	assert(NULL != m);

	if (m != NULL) {
		dlinkDelete(m, &unknown_list);
		dlinkAdd(client_p, m, &serv_list);
	} else {
		sendto_realops_flags(UMODE_ALL, L_ADMIN,
				     "Tried to register (%s) server but it was already registered!?!",
				     host);
		exit_client(client_p, client_p, client_p,
		"Tried to register server but it was already registered?!?");
	}

	Count.myserver++;

	dlinkAddTail(client_p, make_dlink_node(), &global_serv_list);
	hash_add_client(client_p);

	/* doesnt duplicate client_p->serv if allocated this struct already */
	make_server(client_p);
	strlcpy(client_p->serv->up, me.name, sizeof(client_p->serv->up));

	/* fixing eob timings.. -gnp */
	client_p->firsttime = CurrentTime;

	/* Show the real host/IP to admins */
	sendto_realops_flags(UMODE_ALL, L_ADMIN,
			     "Link with %s established, protocol version %d",
			     inpath_ip, client_p->protocol);

	/* Now show the masked hostname/IP to opers */
	sendto_realops_flags(UMODE_ALL, L_OPER,
			     "Link with %s established, protocol version %d",
			     inpath, client_p->protocol);

	ilog(L_NOTICE, "Link with %s established, protocol version %d",
	     inpath_ip, client_p->protocol);

	client_p->serv->sconf = conf;

	fd_note(client_p->localClient->fd, "Server: %s", client_p->name);

	/*
	 * Old sendto_serv_but_one() call removed because we now * need to
	 * send different names to different servers * (domain name matching)
	 * Send new server to other servers.
	 */
	DLINK_FOREACH(ptr, serv_list.head) {
		target_p = ptr->data;

		if (target_p == client_p)
			continue;

		if ((conf = target_p->serv->sconf) && match(my_name_for_link(conf), client_p->name))
			continue;

		if (target_p->protocol > PROTO_MIN && HasID(client_p))
			sendto_one(target_p, ":%s SID %s 2 %s :%s%s",
				   me.id, client_p->name, client_p->id,
			  IsHidden(client_p) ? "(H) " : "", client_p->info);
		else
			sendto_one(target_p, ":%s SERVER %s 2 :%s%s",
				   me.name, client_p->name,
			  IsHidden(client_p) ? "(H) " : "", client_p->info);
	}

	/*
	 * Pass on my client information to the new server *
	 * 
	 * First, pass only servers (idea is that if the link gets * cancelled
	 * beacause the server was already there, * there are no NICK's to be
	 * cancelled...). Of course, * if cancellation occurs, all this info
	 * is sent anyway, * and I guess the link dies when a read is
	 * attempted...? --msa *
	 * 
	 * Note: Link cancellation to occur at this point means * that at
	 * least two servers from my fragment are building * up connection
	 * this other fragment at the same time, it's * a race condition, not
	 * the normal way of operation... *
	 * 
	 * ALSO NOTE: using the get_client_name for server names-- *    see
	 * previous *WARNING*!!! (Also, original inpath *    is destroyed...)
	 */

	conf = client_p->serv->sconf;

	DLINK_FOREACH(ptr, global_serv_list.head) {
		target_p = ptr->data;

		/* target_p->from == target_p for target_p == client_p */
		if (target_p->from == client_p)
			continue;

		if (match(my_name_for_link(conf), target_p->name))
			continue;

		if (client_p->protocol > PROTO_MIN) {
			struct Client  *up;

			if ((up = find_server(target_p->serv->up)) != NULL
			    && HasID(target_p))
				sendto_one(client_p, ":%s SID %s %d %s :%s%s", ID(up),
					   target_p->name, target_p->hopcount + 1,
					   target_p->id, IsHidden(target_p) ? "(H) " : "",
					   target_p->info);
			else if (up != NULL)	/* introducing non-ts6
						 * server linked through
						 * ts6 server */
				sendto_one(client_p, ":%s SERVER %s %d :%s%s",
					   ID(up), target_p->name, target_p->hopcount + 1,
					   IsHidden(target_p) ? "(H) " : "",
					   target_p->info);
			else	/* introducing non-ts6 server linked
				 * through non-ts6 server */
				sendto_one(client_p, ":%s SERVER %s %d :%s%s",
					 target_p->serv->up, target_p->name,
					   target_p->hopcount + 1,
				   IsHidden(target_p) ? "(H) " : "",
					   target_p->info);
		} else
			sendto_one(client_p, ":%s SERVER %s %d :%s%s",
				 target_p->serv->up, target_p->name,
				   target_p->hopcount + 1, IsHidden(target_p) ? "(H) " : "",
				   target_p->info);
	}

	if ((ServerInfo.hub == 0) && MyConnect(client_p))
		uplink = client_p;

	server_burst(client_p);
	return (0);
}

/*
 * server_burst()
 * 
 * inputs       - struct Client pointer server 
 * output       - none 
 * side effects - send a server burst
 * bugs		- still too long
 */
static void
server_burst(struct Client *client_p)
{
	/*
	 * Send it in the shortened format with the TS, if * it's a TS
	 * server; walk the list of channels, sending * all the nicks that
	 * haven't been sent yet for each * channel, then send the channel
	 * itself -- it's less * obvious than sending all nicks first, but on
	 * the * receiving side memory will be allocated more nicely * saving
	 * a few seconds in the handling of a split * -orabidoo
	 */

	burst_all(client_p);

	/* EOB stuff is now in burst_all */
	/* Always send a PING after connect burst is done */
	sendto_one(client_p, "PING :%s", me.name);
}

/*
 * burst_all()
 * 
 * inputs	- pointer to server to send burst to output	- NONE side effects -
 * complete burst of channels/nicks is sent to client_p
 */
static void
burst_all(struct Client *client_p)
{
	struct Client  *target_p;
	struct Channel *chptr;
	struct hook_burst_channel hinfo;
	dlink_node     *gptr;
	dlink_node     *ptr;

	DLINK_FOREACH(ptr, global_client_list.head) {
		target_p = ptr->data;

		if (target_p->from != client_p)
			sendnick_TS(client_p, target_p);
	}

	DLINK_FOREACH(gptr, global_channel_list.head) {
		chptr = gptr->data;

		if (dlink_list_length(&chptr->members) != 0) {
			send_channel_modes(client_p, chptr);
			hinfo.chptr = chptr;
			hinfo.client = client_p;
			hook_call_event("burst_channel", &hinfo);
		}
	}

	sendto_one(client_p, ":%s EOB", ID_or_name(&me, client_p));
}

/*
 * set_autoconn()
 * 
 * inputs       - struct Client pointer to oper requesting change 
 * output       - none 
 * side effects - set autoconnect mode
 */
void
set_autoconn(struct Client *source_p, const char *name, int newval)
{
	struct ConfItem *conf;
	struct AccessItem *aconf;

	if (name != NULL) {
		conf = find_exact_name_conf(SERVER_TYPE, name, NULL, NULL);
		if (conf != NULL) {
			aconf = (struct AccessItem *)map_to_conf(conf);
			if (newval)
				SetConfAllowAutoConn(aconf);
			else
				ClearConfAllowAutoConn(aconf);

			sendto_realops_flags(UMODE_ALL, L_ALL,
				     "%s has changed AUTOCONN for %s to %i",
					     source_p->name, name, newval);
			sendto_one(source_p,
			  ":%s NOTICE %s :AUTOCONN for %s is now set to %i",
				   me.name, source_p->name, name, newval);
		} else {
			sendto_one(source_p, ":%s NOTICE %s :Can't find %s",
				   me.name, source_p->name, name);
		}
	} else {
		sendto_one(source_p, ":%s NOTICE %s :Please specify a server name!",
			   me.name, source_p->name);
	}
}

void
initServerMask(void)
{
	freeMask = 0xFFFFFFFFUL;
}

/*
 * nextFreeMask()
 * 
 * inputs	- NONE output	- unsigned long next unused mask for use in LL side
 * effects	-
 */
unsigned long
nextFreeMask(void)
{
	int		i;
	unsigned long	mask = 1;

	for (i = 0; i < 32; i++) {
		if (mask & freeMask) {
			freeMask &= ~mask;
			return (mask);
		}
		mask <<= 1;
	}

	return (0L);		/* watch this special case ... */
}

/*
 * New server connection code Based upon the stuff floating about in s_bsd.c
 * -- adrian
 */

/*
 * serv_connect() - initiate a server connection
 * 
 * inputs	- pointer to conf - pointer to client doing the connect output	-
 * side effects	-
 * 
 * This code initiates a connection to a server. It first checks to make sure
 * the given server exists. If this is the case, it creates a socket, creates
 * a client, saves the socket information in the client, and initiates a
 * connection to the server through comm_connect_tcp(). The completion of
 * this goes through serv_completed_connection().
 * 
 * We return 1 if the connection is attempted, since we don't know whether it
 * suceeded or not, and 0 if it fails in here somewhere.
 */
int
serv_connect(struct AccessItem *aconf, struct Client *by)
{
	struct ConfItem *conf;
	struct Client  *client_p;
	int		fd;
	char		buf       [HOSTIPLEN];

	/* conversion structs */
	struct sockaddr_in *v4;
	/* Make sure aconf is useful */
	assert(aconf != NULL);

	if (aconf == NULL)
		return (0);

	/* XXX should be passing struct ConfItem in the first place */
	conf = unmap_conf_item(aconf);

	/* log */
	irc_getnameinfo((struct sockaddr *)&aconf->ipnum, aconf->ipnum.ss_len,
			buf, HOSTIPLEN, NULL, 0, NI_NUMERICHOST);
	ilog(L_NOTICE, "Connect to %s[%s] @%s", aconf->user, aconf->host, buf);

	/* Still processing a DNS lookup? -> exit */
	if (aconf->dns_query != NULL) {
		sendto_realops_flags(UMODE_ALL, L_OPER,
				     "Error connecting to %s: Error during DNS lookup", conf->name);
		return (0);
	}
	/*
	 * Make sure this server isn't already connected Note: aconf should
	 * ALWAYS be a valid C: line
	 */
	if ((client_p = find_server(conf->name)) != NULL) {
		sendto_realops_flags(UMODE_ALL, L_ADMIN,
				     "Server %s already present from %s",
			    conf->name, get_client_name(client_p, SHOW_IP));
		sendto_realops_flags(UMODE_ALL, L_OPER,
				     "Server %s already present from %s",
			    conf->name, get_client_name(client_p, MASK_IP));
		if (by && IsPerson(by) && !MyClient(by))
			sendto_one(by, ":%s NOTICE %s :Server %s already present from %s",
				   me.name, by->name, conf->name,
				   get_client_name(client_p, MASK_IP));
		return (0);
	}
	/* create a socket for the server connection */
	if ((fd = comm_open(aconf->ipnum.ss.ss_family, SOCK_STREAM, 0, NULL)) < 0) {
		/* Eek, failure to create the socket */
		report_error(L_ALL, "opening stream socket to %s: %s", conf->name, errno);
		return (0);
	}
	/* servernames are always guaranteed under HOSTLEN chars */
	fd_note(fd, "Server: %s", conf->name);

	/* Create a local client */
	client_p = make_client(NULL);

	/* Copy in the server, hostname, fd */
	strlcpy(client_p->name, conf->name, sizeof(client_p->name));
	strlcpy(client_p->host, aconf->host, sizeof(client_p->host));
	/* We already converted the ip once, so lets use it - stu */
	strlcpy(client_p->localClient->sockhost, buf, HOSTIPLEN);
	client_p->localClient->fd = fd;

	/*
	 * Set up the initial server evilness, ripped straight from
	 * connect_server(), so don't blame me for it being evil. -- adrian
	 */
	if (!set_non_blocking(client_p->localClient->fd)) {
		report_error(L_ADMIN, NONB_ERROR_MSG, get_client_name(client_p, SHOW_IP), errno);
		report_error(L_OPER, NONB_ERROR_MSG, get_client_name(client_p, MASK_IP), errno);
	}
	if (!set_sock_buffers(client_p->localClient->fd, READBUF_SIZE)) {
		report_error(L_ADMIN, SETBUF_ERROR_MSG, get_client_name(client_p, SHOW_IP), errno);
		report_error(L_OPER, SETBUF_ERROR_MSG, get_client_name(client_p, MASK_IP), errno);
	}
	/*
	 * Attach config entries to client here rather than in
	 * serv_connect_callback(). This to avoid null pointer references.
	 */
	if (!attach_connect_block(client_p, conf->name, aconf->host)) {
		sendto_realops_flags(UMODE_ALL, L_ALL,
			"Host %s is not enabled for connecting:no C/N-line",
				     conf->name);
		if (by && IsPerson(by) && !MyClient(by))
			sendto_one(by, ":%s NOTICE %s :Connect to host %s failed.",
				   me.name, by->name, client_p->name);
		detach_all_confs(client_p);
		return (0);
	}
	/*
	 * at this point we have a connection in progress and C/N lines
	 * attached to the client, the socket info should be saved in the
	 * client and it should either be resolved or have a valid address.
	 * 
	 * The socket has been connected or connect is in progress.
	 */
	make_server(client_p);

	if (by && IsPerson(by))
		strlcpy(client_p->serv->by, by->name, sizeof(client_p->serv->by));
	else
		strlcpy(client_p->serv->by, "AutoConn.", sizeof(client_p->serv->by));

	strlcpy(client_p->serv->up, me.name, sizeof(client_p->serv->up));
	SetConnecting(client_p);
	dlinkAdd(client_p, &client_p->node, &global_client_list);
	/* from def_fam */
	client_p->localClient->aftype = aconf->aftype;

	/* Now, initiate the connection */
	/*
	 * XXX assume that a non 0 type means a specific bind address for
	 * this connect.
	 */
	switch (aconf->aftype) {
	case AF_INET:
		v4 = (struct sockaddr_in *)&aconf->my_ipnum;
		if (v4->sin_addr.s_addr) {
			struct irc_ssaddr ipn;
			memset(&ipn, 0, sizeof(struct irc_ssaddr));
			ipn.ss.ss_family = AF_INET;
			ipn.ss_port = 0;
			memcpy(&ipn, &aconf->my_ipnum, sizeof(struct irc_ssaddr));
			comm_connect_tcp(client_p->localClient->fd, aconf->host, aconf->port,
					 (struct sockaddr *)&ipn, ipn.ss_len,
			     serv_connect_callback, client_p, aconf->aftype,
					 CONNECTTIMEOUT);
		} else if (ServerInfo.specific_ipv4_vhost) {
			struct irc_ssaddr ipn;
			memset(&ipn, 0, sizeof(struct irc_ssaddr));
			ipn.ss.ss_family = AF_INET;
			ipn.ss_port = 0;
			memcpy(&ipn, &ServerInfo.ip, sizeof(struct irc_ssaddr));
			comm_connect_tcp(client_p->localClient->fd, aconf->host, aconf->port,
					 (struct sockaddr *)&ipn, ipn.ss_len,
			     serv_connect_callback, client_p, aconf->aftype,
					 CONNECTTIMEOUT);
		} else
			comm_connect_tcp(client_p->localClient->fd, aconf->host, aconf->port,
					 NULL, 0, serv_connect_callback, client_p, aconf->aftype,
					 CONNECTTIMEOUT);
		break;
#ifdef IPV6
	case AF_INET6:
		if (ServerInfo.specific_ipv6_vhost) {
			struct irc_ssaddr ipn;
			memset(&ipn, 0, sizeof(struct irc_ssaddr));
			ipn.ss.ss_family = AF_INET6;
			ipn.ss_port = 0;
			memcpy(&ipn, &ServerInfo.ip6, sizeof(struct irc_ssaddr));
			comm_connect_tcp(client_p->localClient->fd, aconf->host, aconf->port,
					 (struct sockaddr *)&ipn, ipn.ss_len,
			     serv_connect_callback, client_p, aconf->aftype,
					 CONNECTTIMEOUT);
		} else
			comm_connect_tcp(client_p->localClient->fd, aconf->host, aconf->port,
					 NULL, 0, serv_connect_callback, client_p, aconf->aftype,
					 CONNECTTIMEOUT);
#endif
	}
	return (1);
}

#if 0
static char
*sslerr_to_string(int error)
{
        switch (error)
        {
        case SSL_ERROR_NONE:
                return "No error";
                break;
        case SSL_ERROR_SSL:
                return "Internal OpenSSL error or protocol error";
                break;
        case SSL_ERROR_WANT_READ:
                return "OpenSSL functions requested a read()";
                break;
        case SSL_ERROR_WANT_WRITE:
                return "OpenSSL functions requested a write()";
                break;
        case SSL_ERROR_WANT_X509_LOOKUP:
                return "OpenSSL requested a X509 lookup which didn`t arrive";
                break;
        case SSL_ERROR_SYSCALL:
                return "Underlying syscall error";
                break;
        case SSL_ERROR_ZERO_RETURN:
                return "Underlying socket operation returned zero";
                break;
        case SSL_ERROR_WANT_CONNECT:
                return "OpenSSL functions wanted a connect()";
                break;
        default:
                return "Unknown OpenSSL error (huh?)";
        }
}
#endif

/*
 * serv_connect_callback() - complete a server connection.
 * 
 * This routine is called after the server connection attempt has completed. If
 * unsucessful, an error is sent to ops and the client is closed. If
 * sucessful, it goes through the initialisation/check procedures, the
 * capabilities are sent, and the socket is then marked for reading.
 */
static void
serv_connect_callback(int fd, int status, void *data)
{
	struct Client  *client_p = data;
	struct ConfItem *conf = NULL;
	struct AccessItem *aconf = NULL;

	/* First, make sure its a real client! */
	assert(client_p != NULL);
	assert(client_p->localClient->fd == fd);

	set_no_delay(fd);

	/* Next, for backward purposes, record the ip of the server */
	memcpy(&client_p->localClient->ip, &fd_table[fd].connect.hostaddr,
	       sizeof(struct irc_ssaddr));
	/* Check the status */
	if (status != COMM_OK) {
		/*
		 * We have an error, so report it and quit Admins get to see
		 * any IP, mere opers don't *sigh*
		 */
		sendto_realops_flags(UMODE_ALL, L_ADMIN,
			   "Error connecting to %s[%s]: %s", client_p->name,
				     client_p->host, comm_errstr(status));

		sendto_realops_flags(UMODE_ALL, L_OPER,
				     "Error connecting to %s: %s",
				     client_p->name, comm_errstr(status));

		/*
		 * If a fd goes bad, call dead_link() the socket is no longer
		 * valid for reading or writing.
		 */
		dead_link_on_write(client_p, 0);
		return;
	}
	/* COMM_OK, so continue the connection procedure */
	/* Get the C/N lines */
	conf = find_conf_name(&client_p->localClient->confs, client_p->name, SERVER_TYPE);
	if (conf == NULL) {
		sendto_realops_flags(UMODE_ALL, L_ADMIN,
		    "Lost connect{} block for %s", get_client_name(client_p,
								   HIDE_IP));
		sendto_realops_flags(UMODE_ALL, L_OPER, "Lost connect{} block for %s",
				     get_client_name(client_p, MASK_IP));

		exit_client(client_p, client_p, &me, "Lost connect{} block");
		return;
	}
	aconf = (struct AccessItem *)map_to_conf(conf);
#ifdef HAVE_LIBCRYPTO
	set_blocking(client_p->localClient->fd);
	if (aconf->flags & CONF_FLAGS_CRYPTLINK)
	{
		int i;

		fde_t *F = &fd_table[client_p->localClient->fd];
		if ((F->ssl = SSL_new(ServerInfo.ctx2)) == NULL)
		{
			close_connection(client_p);
			free_client(client_p);
		}
		SetSSL(client_p);

		SSL_set_fd(F->ssl, client_p->localClient->fd);

		SSL_connect(F->ssl);

		/* cryptlink kludge: Send about 10 blank lines down
		 * the pipe. This way SSL is initialized without any
		 * problematic issues, such as not being able to read.
		 */
		for (i = 0; i < 10; i++)
			sendto_one(client_p, "PING %d", i);
	}
	set_non_blocking(client_p->localClient->fd);
#endif

	/* Next, send the initial handshake */
	SetHandshake(client_p);

	/* jdc -- Check and send spasswd, not passwd. */
	if (!EmptyString(aconf->spasswd))
		sendto_one(client_p, "PASS %s TS %d %s", aconf->spasswd, TS_CURRENT, me.id);

	sendto_one(client_p, "SERVER %s 1 %d %s :%s",
		   my_name_for_link(conf), PROTO_CUR,
		   aconf->flags & CONF_FLAGS_COMPRESSED ? "ZIP" : "", 
		   me.info);

	sendto_one(client_p, "SVINFO %d 3 0 :%lu",
		   TS_CURRENT, RealTime);

	/*
	 * If we've been marked dead because a send failed, just exit here
	 * now and save everyone the trouble of us ever existing.
	 */
	if (IsDead(client_p)) {
		sendto_realops_flags(UMODE_ALL, L_ADMIN,
				     "%s[%s] went dead during handshake",
				     client_p->name, client_p->host);
		sendto_realops_flags(UMODE_ALL, L_OPER,
			   "%s went dead during handshake", client_p->name);
		return;
	}
	/* don't move to serv_list yet -- we haven't sent a burst! */
	/* If we get here, we're ok, so lets start reading some data */
	comm_setselect(fd, FDLIST_SERVER, COMM_SELECT_READ, read_packet, client_p, 0);
}
